"""
Dotfiles stage for kernel lifecycle.

This stage handles processing and installation of dotfiles in containers.
"""

import asyncio
from dataclasses import dataclass
from pathlib import Path
from typing import Optional, override

from ai.backend.common.stage.types import ArgsSpecGenerator, Provisioner, ProvisionStage

from .utils import ChownUtil, PathOwnerDeterminer


@dataclass
class DotfileInput:
    path: str
    data: str
    perm: str  # Octal permission string like "644"


@dataclass
class DirOwnPermission:
    path: str
    perm: str
    uid: Optional[int]
    gid: Optional[int]


@dataclass
class DotfileProcResult:
    path: str
    data: str
    dir_own_permissions: list[DirOwnPermission]


@dataclass
class AgentConfig:
    kernel_features: frozenset[str]
    kernel_uid: int
    kernel_gid: int


@dataclass
class DotfilesSpec:
    dotfiles: list[DotfileInput]
    scratch_dir: Path
    work_dir: Path

    # Override UID/GID settings
    uid_override: Optional[int]
    gid_override: Optional[int]

    agent_config: AgentConfig


class DotfilesSpecGenerator(ArgsSpecGenerator[DotfilesSpec]):
    pass


@dataclass
class DotfilesResult:
    processed_dotfiles: list[DotfileProcResult]


class DotfilesProvisioner(Provisioner[DotfilesSpec, DotfilesResult]):
    """
    Provisioner for dotfiles processing.

    Processes and installs dotfiles from internal_data into the container filesystem.
    """

    @property
    @override
    def name(self) -> str:
        return "docker-dotfiles"

    @override
    async def setup(self, spec: DotfilesSpec) -> DotfilesResult:
        loop = asyncio.get_running_loop()

        result = await loop.run_in_executor(None, self._process_dotfiles, spec)
        return DotfilesResult(processed_dotfiles=result)

    def _process_dotfiles(self, spec: DotfilesSpec) -> list[DotfileProcResult]:
        if not spec.dotfiles:
            return []
        chown = ChownUtil()
        owner_determiner = PathOwnerDeterminer.by_kernel_features(
            spec.agent_config.kernel_uid,
            spec.agent_config.kernel_gid,
            spec.agent_config.kernel_features,
        )
        determined_uid: Optional[int] = None
        determined_gid: Optional[int] = None
        processed_dotfiles: list[DotfileProcResult] = []
        for dotfile in spec.dotfiles:
            file_path = self._resolve_dotfile_path(spec, dotfile.path)

            # Create parent directories
            file_path.parent.mkdir(parents=True, exist_ok=True)

            # Write file content
            self._write_dotfile(dotfile.data, file_path)

            # Set permissions and ownership
            target_paths = self._get_parent_dirs(spec.work_dir, file_path, dotfile.perm)

            uid, gid = owner_determiner.determine(
                file_path, uid_override=spec.uid_override, gid_override=spec.gid_override
            )
            chown.chown_paths(target_paths, uid, gid)
            if determined_uid is None:
                determined_uid = uid
            if determined_gid is None:
                determined_gid = gid

            processed_dotfiles.append(
                DotfileProcResult(
                    path=str(file_path),
                    data=dotfile.data,
                    dir_own_permissions=[
                        DirOwnPermission(
                            path=str(p),
                            perm=dotfile.perm,
                            uid=determined_uid,
                            gid=determined_gid,
                        )
                        for p in target_paths
                    ],
                )
            )

        return processed_dotfiles

    def _resolve_dotfile_path(self, spec: DotfilesSpec, dotfile_path: str) -> Path:
        """
        Determine the file path for a dotfile based on its path and the scratch directory.
        """
        if dotfile_path.startswith("/"):
            if dotfile_path.startswith("/home/"):
                # Handle paths under /home/ by redirecting to scratch dir
                path_arr = dotfile_path.split("/")
                file_path = spec.scratch_dir / "/".join(path_arr[2:])
            else:
                # Absolute paths outside /home/
                file_path = Path(dotfile_path)
        else:
            # Relative paths go under work_dir
            file_path = spec.work_dir / dotfile_path
        return file_path

    def _write_dotfile(self, dotfile_content: str, dotfile_path: Path) -> None:
        if not dotfile_content.endswith("\n"):
            dotfile_content += "\n"
        dotfile_path.write_text(dotfile_content)

    def _get_parent_dirs(self, work_dir: Path, base_path: Path, perm: str) -> list[Path]:
        """
        Get all parent directories of a given path.
        """
        # Set permissions and ownership
        tmp = Path(base_path)
        tmp_paths: list[Path] = []

        # Set permissions on file and all parent directories up to work_dir
        while tmp != work_dir:
            tmp.chmod(int(perm, 8))
            tmp_paths.append(tmp)
            tmp = tmp.parent
        return tmp_paths

    @override
    async def teardown(self, resource: DotfilesResult) -> None:
        # Dotfiles are cleaned up with scratch directory
        pass


class DotfilesStage(ProvisionStage[DotfilesSpec, DotfilesResult]):
    """
    Stage for processing and installing dotfiles in kernel containers.
    """

    pass



"""
A module to collect various performance metrics of Docker containers.

Reference: https://www.datadoghq.com/blog/how-to-collect-docker-metrics/
"""

import asyncio
import enum
import logging
import sys
import time
import uuid
from decimal import Decimal, DecimalException
from typing import (
    TYPE_CHECKING,
    Callable,
    FrozenSet,
    List,
    Mapping,
    MutableMapping,
    Optional,
    Sequence,
    Set,
    Tuple,
    cast,
)

import aiodocker
import attrs

from ai.backend.common import msgpack
from ai.backend.common.identity import is_containerized
from ai.backend.common.metrics.metric import StageObserver
from ai.backend.common.types import (
    PID,
    ContainerId,
    DeviceId,
    KernelId,
    MetricKey,
    MetricValue,
    MovingStatValue,
    SessionId,
)
from ai.backend.logging import BraceStyleAdapter

from .metrics.metric import UtilizationMetricObserver
from .metrics.types import (
    CAPACITY_METRIC_KEY,
    CURRENT_METRIC_KEY,
    PCT_METRIC_KEY,
    FlattenedDeviceMetric,
    FlattenedKernelMetric,
)
from .utils import remove_exponent

if TYPE_CHECKING:
    from .agent import AbstractAgent
    from .kernel import AbstractKernel

__all__ = (
    "StatContext",
    "StatModes",
    "MetricTypes",
    "NodeMeasurement",
    "ContainerMeasurement",
    "Measurement",
)

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


def check_cgroup_available():
    """
    Check if the host OS provides cgroups.
    """
    return not is_containerized() and sys.platform.startswith("linux")


class StatModes(enum.StrEnum):
    CGROUP = "cgroup"
    DOCKER = "docker"

    @staticmethod
    def get_preferred_mode():
        """
        Returns the most preferred statistics collector type for the host OS.
        """
        if check_cgroup_available():
            return StatModes.CGROUP
        return StatModes.DOCKER


class MetricTypes(enum.Enum):
    """
    Specifies the type of a metric value.

    Currently this DOES NOT affect calculation and processing of the metric,
    but serves as a metadata for code readers.
    The actual calculation and formatting is controlled by :meth:`Metric.current_hook()`,
    :attr:`Metric.unit_hint` and :attr:`Metric.stats_filter`.
    """

    GAUGE = 0
    """
    Represents an instantly measured occupancy value.
    (e.g., used space as bytes, occupied amount as the number of items or a bandwidth)
    """
    USAGE = 0
    """
    This is same to GAUGE, but just kept for backward compatibility of compute plugins.
    """
    RATE = 1
    """
    Represents a rate of changes calculated from underlying gauge/accumulation values
    (e.g., I/O bps calculated from RX/TX accum.bytes)
    """
    UTILIZATION = 2
    """
    Represents a ratio of resource occupation time per each measurement interval
    (e.g., CPU utilization)
    """
    ACCUMULATION = 3
    """
    Represents an accumulated value
    (e.g., total number of events, total period of occupation)
    """


@attrs.define(auto_attribs=True, slots=True)
class Measurement:
    value: Decimal
    capacity: Optional[Decimal] = None


@attrs.define(auto_attribs=True, slots=True)
class NodeMeasurement:
    """
    Collection of per-node and per-agent statistics for a specific metric.
    """

    # 2-tuple of Decimals mean raw values for (usage, available)
    # Percent values are calculated from them.
    key: MetricKey
    type: MetricTypes
    per_node: Measurement
    per_device: Mapping[DeviceId, Measurement] = attrs.Factory(dict)
    stats_filter: FrozenSet[str] = attrs.Factory(frozenset)
    current_hook: Optional[Callable[["Metric"], Decimal]] = None
    unit_hint: str = "count"


@attrs.define(auto_attribs=True, slots=True)
class ContainerMeasurement:
    """
    Collection of per-container statistics for a specific metric.
    """

    key: MetricKey
    type: MetricTypes
    per_container: Mapping[str, Measurement] = attrs.Factory(dict)
    stats_filter: FrozenSet[str] = attrs.Factory(frozenset)
    current_hook: Optional[Callable[["Metric"], Decimal]] = None
    unit_hint: str = "count"


@attrs.define(auto_attribs=True, slots=True)
class ProcessMeasurement:
    """
    Collection of per-process statistics for a specific metric.
    """

    key: MetricKey
    type: MetricTypes
    per_process: Mapping[int, Measurement] = attrs.Factory(dict)
    stats_filter: FrozenSet[str] = attrs.Factory(frozenset)
    current_hook: Optional[Callable[["Metric"], Decimal]] = None
    unit_hint: str = "count"


def _to_serializable_value(value: Decimal, *, exponent: Decimal = Decimal("0.000")) -> str:
    """
    Convert a Decimal value to a string without scientific notation.
    """
    try:
        quantized = value.quantize(exponent)
    except DecimalException:
        return str(value)
    try:
        return str(remove_exponent(quantized))
    except DecimalException:
        return str(quantized)


class MovingStatistics:
    __slots__ = (
        "_sum",
        "_count",
        "_min",
        "_max",
        "_last",
    )
    _sum: Decimal
    _count: int
    _min: Decimal
    _max: Decimal
    _last: List[Tuple[Decimal, float]]

    def __init__(self, initial_value: Optional[Decimal] = None):
        self._last = []
        if initial_value is None:
            self._sum = Decimal(0)
            self._min = Decimal("inf")
            self._max = Decimal("-inf")
            self._count = 0
        else:
            self._sum = initial_value
            self._min = initial_value
            self._max = initial_value
            self._count = 1
            point = (initial_value, time.perf_counter())
            self._last.append(point)

    def update(self, value: Decimal):
        self._sum += value
        self._min = min(self._min, value)
        self._max = max(self._max, value)
        self._count += 1
        point = (value, time.perf_counter())
        self._last.append(point)
        # keep only the latest two data points
        if len(self._last) > 2:
            self._last.pop(0)

    @property
    def min(self) -> Decimal:
        return self._min

    @property
    def max(self) -> Decimal:
        return self._max

    @property
    def sum(self) -> Decimal:
        return self._sum

    @property
    def avg(self) -> Decimal:
        return self._sum / self._count

    @property
    def diff(self) -> Decimal:
        if len(self._last) == 2:
            return self._last[-1][0] - self._last[-2][0]
        return Decimal(0)

    @property
    def rate(self) -> Decimal:
        if len(self._last) == 2:
            return (self._last[-1][0] - self._last[-2][0]) / Decimal(
                self._last[-1][1] - self._last[-2][1]
            )
        return Decimal(0)

    def to_serializable_dict(self) -> MovingStatValue:
        return {
            "min": _to_serializable_value(self.min),
            "max": _to_serializable_value(self.max),
            "sum": _to_serializable_value(self.sum),
            "avg": _to_serializable_value(self.avg),
            "diff": _to_serializable_value(self.diff),
            "rate": _to_serializable_value(self.rate),
            "version": 2,
        }

    def __str__(self) -> str:
        return str({
            "min": self.min,
            "max": self.max,
            "sum": self.sum,
            "avg": self.avg,
            "diff": self.diff,
            "rate": self.rate,
        })


@attrs.define(auto_attribs=True, slots=True)
class Metric:
    key: str  # MetricKey
    type: MetricTypes
    unit_hint: str
    stats: MovingStatistics
    stats_filter: FrozenSet[str]
    current: Decimal
    capacity: Optional[Decimal] = None
    current_hook: Optional[Callable[["Metric"], Decimal]] = None

    def update(self, value: Measurement):
        if value.capacity is not None:
            self.capacity = value.capacity
        self.stats.update(value.value)
        self.current = value.value
        if self.current_hook is not None:
            self.current = self.current_hook(self)

    def to_serializable_dict(self) -> MetricValue:
        q_pct = Decimal("0.00")
        return {
            "current": _to_serializable_value(self.current),
            "capacity": (
                _to_serializable_value(self.capacity) if self.capacity is not None else None
            ),
            "pct": (
                _to_serializable_value(
                    (Decimal(self.current) / Decimal(self.capacity) * 100), exponent=q_pct
                )
                if (self.capacity is not None and self.capacity.is_normal() and self.capacity > 0)
                else "0.00"
            ),
            "unit_hint": self.unit_hint,
            **{
                f"stats.{k}": v  # type: ignore
                for k, v in self.stats.to_serializable_dict().items()
                if k in self.stats_filter
            },
        }


class StatContext:
    agent: "AbstractAgent"
    mode: StatModes
    node_metrics: dict[MetricKey, Metric]
    device_metrics: dict[MetricKey, dict[DeviceId, Metric]]
    kernel_metrics: dict[KernelId, dict[MetricKey, Metric]]
    process_metrics: dict[ContainerId, dict[PID, dict[MetricKey, Metric]]]
    _utilization_metric_observer: UtilizationMetricObserver
    _stage_observer: StageObserver

    def __init__(
        self, agent: "AbstractAgent", mode: Optional[StatModes] = None, *, cache_lifespan: int = 120
    ) -> None:
        self.agent = agent
        self.mode = mode if mode is not None else StatModes.get_preferred_mode()
        self.cache_lifespan = cache_lifespan

        self.node_metrics = {}
        self.device_metrics = {}
        self.kernel_metrics = {}
        self.process_metrics = {}

        self._lock = asyncio.Lock()
        self._timestamps: MutableMapping[str, float] = {}
        self._utilization_metric_observer = UtilizationMetricObserver.instance()
        self._stage_observer = StageObserver.instance()

    def update_timestamp(self, timestamp_key: str) -> Tuple[float, float]:
        """
        Update the timestamp for the given key and return a pair of the current timestamp and
        the interval from the last update of the same key.

        If the last timestamp for the given key does not exist, the interval becomes "NaN".

        Intended to be used by compute plugins.
        """
        now = time.perf_counter()
        last = self._timestamps.get(timestamp_key, None)
        self._timestamps[timestamp_key] = now
        if last is None:
            return now, 0.0
        return now, now - last

    def _get_ownership_info_from_kernel(
        self,
        kernel_id: KernelId,
    ) -> tuple[Optional[SessionId], Optional[uuid.UUID], Optional[uuid.UUID]]:
        kernel_obj = self.agent.kernel_registry.get(kernel_id)
        if kernel_obj is not None:
            ownership_data = kernel_obj.ownership_data
            session_id = ownership_data.session_id
            owner_user_id = ownership_data.owner_user_id
            owner_project_id = ownership_data.owner_project_id
        else:
            session_id = None
            owner_user_id = None
            owner_project_id = None
        return session_id, owner_user_id, owner_project_id

    def observe_node_metric(
        self,
        device_id: DeviceId,
        metric_key: MetricKey,
        measure: Measurement,
    ) -> None:
        agent_id = self.agent.id
        value_pairs = [
            (CURRENT_METRIC_KEY, str(measure.value)),
        ]
        if measure.capacity is not None:
            value_pairs.append((CAPACITY_METRIC_KEY, str(measure.capacity)))
        self._utilization_metric_observer.observe_device_metric(
            metric=FlattenedDeviceMetric(
                agent_id=agent_id,
                device_id=device_id,
                key=metric_key,
                value_pairs=value_pairs,
            )
        )

    async def remove_kernel_metric(self, kernel_id: KernelId) -> None:
        log.info("Removing metrics for kernel {}", kernel_id)
        known_metrics = self.kernel_metrics.get(kernel_id)
        log.debug("Known metrics for kernel {}: {}", kernel_id, known_metrics)
        if known_metrics is None:
            return
        metric_keys = list(known_metrics.keys())
        agent_id = self.agent.id
        session_id, owner_user_id, project_id = self._get_ownership_info_from_kernel(kernel_id)
        await self._utilization_metric_observer.lazy_remove_container_metric(
            agent_id=agent_id,
            kernel_id=kernel_id,
            session_id=session_id,
            owner_user_id=owner_user_id,
            project_id=project_id,
            keys=metric_keys,
        )

    async def collect_node_stat(self) -> None:
        """
        Collect the per-node, per-device, and per-container statistics.

        Intended to be used by the agent.
        """
        self._stage_observer.observe_stage(
            stage="before_lock",
            upper_layer="collect_node_stat",
        )
        async with self._lock:
            # Here we use asyncio.gather() instead of aiotools.TaskGroup
            # to keep methods of other plugins running when a plugin raises an error
            # instead of cancelling them.
            _tasks: list[asyncio.Task[Sequence[NodeMeasurement]]] = []
            for computer in self.agent.computers.values():
                _tasks.append(asyncio.create_task(computer.instance.gather_node_measures(self)))
            self._stage_observer.observe_stage(
                stage="before_gather_measures",
                upper_layer="collect_node_stat",
            )
            results = await asyncio.gather(*_tasks, return_exceptions=True)
            self._stage_observer.observe_stage(
                stage="before_observe",
                upper_layer="collect_node_stat",
            )
            for result in results:
                if isinstance(result, BaseException):
                    log.error("collect_node_stat(): gather_node_measures() error", exc_info=result)
                    continue
                for node_measure in result:
                    metric_key = node_measure.key
                    # update node metric
                    if metric_key not in self.node_metrics:
                        self.node_metrics[metric_key] = Metric(
                            metric_key,
                            node_measure.type,
                            current=node_measure.per_node.value,
                            capacity=node_measure.per_node.capacity,
                            unit_hint=node_measure.unit_hint,
                            stats=MovingStatistics(node_measure.per_node.value),
                            stats_filter=frozenset(node_measure.stats_filter),
                            current_hook=node_measure.current_hook,
                        )
                    else:
                        self.node_metrics[metric_key].update(node_measure.per_node)
                    # update per-device metric
                    for dev_id, measure in node_measure.per_device.items():
                        self.observe_node_metric(
                            device_id=dev_id,
                            metric_key=metric_key,
                            measure=measure,
                        )
                        if metric_key not in self.device_metrics:
                            self.device_metrics[metric_key] = {}
                        if dev_id not in self.device_metrics[metric_key]:
                            self.device_metrics[metric_key][dev_id] = Metric(
                                metric_key,
                                node_measure.type,
                                current=measure.value,
                                capacity=measure.capacity,
                                unit_hint=node_measure.unit_hint,
                                stats=MovingStatistics(measure.value),
                                stats_filter=frozenset(node_measure.stats_filter),
                                current_hook=node_measure.current_hook,
                            )
                        else:
                            self.device_metrics[metric_key][dev_id].update(measure)
        agent_id = self.agent.id
        device_metrics: dict[MetricKey, dict[DeviceId, MetricValue]] = {}
        flattened_metrics: list[FlattenedDeviceMetric] = []
        for metric_key, per_device in self.device_metrics.items():
            if metric_key not in device_metrics:
                device_metrics[metric_key] = {}
            for device_id, obj in per_device.items():
                try:
                    metric_value = obj.to_serializable_dict()
                except ValueError:
                    log.warning(
                        "Failed to serialize metric (Device Id: {}, {})", device_id, str(obj.stats)
                    )
                    continue
                device_metrics[metric_key][device_id] = metric_value
                value_pairs = [
                    (CURRENT_METRIC_KEY, metric_value["current"]),
                    (PCT_METRIC_KEY, metric_value["pct"]),
                ]
                if (capacity := metric_value["capacity"]) is not None:
                    value_pairs.append((CAPACITY_METRIC_KEY, capacity))
                flattened_metrics.append(
                    FlattenedDeviceMetric(
                        agent_id,
                        device_id,
                        metric_key,
                        value_pairs,
                    )
                )

        # push to the Redis server
        node_metrics: dict[MetricKey, MetricValue] = {}
        for key, obj in self.node_metrics.items():
            try:
                node_metrics[key] = obj.to_serializable_dict()
            except ValueError:
                log.warning(
                    "Failed to serialize node metric (Metric key: {}, {})", key, str(obj.stats)
                )
                continue

        redis_agent_updates = {
            "node": node_metrics,
            "devices": device_metrics,
        }
        if self.agent.local_config.debug.log_stats:
            log.debug(
                "stats: node_updates: {0}: {1}",
                self.agent.id,
                redis_agent_updates["node"],
            )
        serialized_agent_updates = msgpack.packb(redis_agent_updates)

        self._stage_observer.observe_stage(
            stage="before_report_to_redis",
            upper_layer="collect_node_stat",
        )

        # Use ValkeyStatClient set method with expiration
        agent_id = self.agent.id
        await self.agent.valkey_stat_client.set(
            agent_id, serialized_agent_updates, expire_sec=self.cache_lifespan
        )

    def observe_container_metric(
        self,
        kernel_id: KernelId,
        metric_key: MetricKey,
        measure: Measurement,
    ) -> None:
        agent_id = self.agent.id
        session_id, owner_user_id, project_id = self._get_ownership_info_from_kernel(kernel_id)
        value_pairs = [
            (CURRENT_METRIC_KEY, str(measure.value)),
        ]
        if measure.capacity is not None:
            value_pairs.append((CAPACITY_METRIC_KEY, str(measure.capacity)))
        self._utilization_metric_observer.observe_container_metric(
            metric=FlattenedKernelMetric(
                agent_id=agent_id,
                kernel_id=kernel_id,
                session_id=session_id,
                owner_user_id=owner_user_id,
                project_id=project_id,
                key=metric_key,
                value_pairs=value_pairs,
            )
        )

    async def collect_container_stat(
        self,
        container_ids: Sequence[ContainerId],
    ) -> None:
        """
        Collect the per-container statistics only,

        Intended to be used by the agent and triggered by container cgroup synchronization processes.
        """
        self._stage_observer.observe_stage(
            stage="before_lock",
            upper_layer="collect_container_stat",
        )
        async with self._lock:
            kernel_id_map: dict[ContainerId, KernelId] = {}
            kernel_obj_map: dict[KernelId, AbstractKernel] = {}
            for kid, info in self.agent.kernel_registry.items():
                try:
                    cid = info["container_id"]
                except KeyError:
                    log.warning("collect_container_stat(): no container for kernel {}", kid)
                else:
                    kernel_id_map[ContainerId(cid)] = kid
                    kernel_obj_map[kid] = info
            unused_kernel_ids = set(self.kernel_metrics.keys()) - set(kernel_id_map.values())
            for unused_kernel_id in unused_kernel_ids:
                log.debug("removing kernel_metric for {}", unused_kernel_id)
                self.kernel_metrics.pop(unused_kernel_id, None)

            # Here we use asyncio.gather() instead of aiotools.TaskGroup
            # to keep methods of other plugins running when a plugin raises an error
            # instead of cancelling them.
            _tasks: list[asyncio.Task[Sequence[ContainerMeasurement]]] = []
            kernel_id = None
            for computer in self.agent.computers.values():
                _tasks.append(
                    asyncio.create_task(
                        computer.instance.gather_container_measures(self, container_ids),
                    )
                )
            self._stage_observer.observe_stage(
                stage="before_gather_measures",
                upper_layer="collect_container_stat",
            )
            results = await asyncio.gather(*_tasks, return_exceptions=True)
            updated_kernel_ids: Set[KernelId] = set()
            self._stage_observer.observe_stage(
                stage="before_observe",
                upper_layer="collect_container_stat",
            )
            for result in results:
                if isinstance(result, BaseException):
                    log.error(
                        "collect_container_stat(): gather_container_measures() error",
                        exc_info=result,
                    )
                    continue
                for ctnr_measure in result:
                    assert isinstance(ctnr_measure, ContainerMeasurement)
                    metric_key = ctnr_measure.key
                    # update per-container metric
                    for cid, measure in ctnr_measure.per_container.items():
                        try:
                            kernel_id = kernel_id_map[ContainerId(cid)]
                        except KeyError:
                            continue
                        self.observe_container_metric(
                            kernel_id,
                            metric_key,
                            measure,
                        )
                        updated_kernel_ids.add(kernel_id)
                        if kernel_id not in self.kernel_metrics:
                            self.kernel_metrics[kernel_id] = {}
                        if metric_key not in self.kernel_metrics[kernel_id]:
                            self.kernel_metrics[kernel_id][metric_key] = Metric(
                                metric_key,
                                ctnr_measure.type,
                                current=measure.value,
                                capacity=measure.capacity or measure.value,
                                unit_hint=ctnr_measure.unit_hint,
                                stats=MovingStatistics(measure.value),
                                stats_filter=frozenset(ctnr_measure.stats_filter),
                                current_hook=ctnr_measure.current_hook,
                            )
                        else:
                            self.kernel_metrics[kernel_id][metric_key].update(measure)

        kernel_updates: list[FlattenedKernelMetric] = []
        kernel_serialized_updates: list[tuple[KernelId, bytes]] = []
        agent_id = self.agent.id
        for kernel_id in updated_kernel_ids:
            session_id, owner_user_id, project_id = self._get_ownership_info_from_kernel(kernel_id)
            metrics = self.kernel_metrics[kernel_id]
            serializable_metrics: dict[MetricKey, MetricValue] = {}
            for key, obj in metrics.items():
                try:
                    metric_value = obj.to_serializable_dict()
                except ValueError:
                    log.warning(
                        "Failed to serialize metric (Metric key: {}, {})", key, str(obj.stats)
                    )
                    continue
                serializable_metrics[key] = metric_value
                value_pairs = [
                    (CURRENT_METRIC_KEY, metric_value["current"]),
                    (PCT_METRIC_KEY, metric_value["pct"]),
                ]
                if (capacity := metric_value["capacity"]) is not None:
                    value_pairs.append((CAPACITY_METRIC_KEY, capacity))
                kernel_updates.append(
                    FlattenedKernelMetric(
                        agent_id,
                        kernel_id,
                        session_id,
                        owner_user_id,
                        project_id,
                        key,
                        value_pairs,
                    )
                )
            if self.agent.local_config.debug.log_stats:
                log.debug("kernel_updates: {0}: {1}", kernel_id, serializable_metrics)

            kernel_serialized_updates.append((kernel_id, msgpack.packb(serializable_metrics)))

        self._stage_observer.observe_stage(
            stage="before_report_to_redis",
            upper_layer="collect_container_stat",
        )

        # Use ValkeyStatClient set_multiple_keys for batch operations
        key_value_map = {str(kernel_id): update for kernel_id, update in kernel_serialized_updates}
        if key_value_map:
            await self.agent.valkey_stat_client.set_multiple_keys(key_value_map)

    async def _get_processes(
        self, container_id: ContainerId, docker: aiodocker.Docker
    ) -> list[PID]:
        """
        Get the list of PIDs for the given container ID.
        """
        return_val: list[PID] = []
        try:
            result = await docker._query_json(f"containers/{container_id}/top", method="GET")
            procs = result["Processes"]
        except (KeyError, aiodocker.exceptions.DockerError):
            log.debug(
                "collect_per_container_process_stat(): cannot find container {}", container_id
            )
            return return_val

        for proc in procs:
            try:
                return_val.append(PID(int(proc[1])))
            except (ValueError, KeyError):
                log.debug(
                    "collect_per_container_process_stat(): cannot parse PID from {}",
                    proc,
                )
                continue
        return return_val

    async def collect_per_container_process_stat(
        self,
        container_ids: Sequence[ContainerId],
    ) -> None:
        """
        Collect the per-container process statistics only,

        Intended to be used by the agent.
        """
        # FIXME: support Docker Desktop backend (#1230)
        if sys.platform == "darwin":
            return

        self._stage_observer.observe_stage(
            stage="before_lock",
            upper_layer="collect_per_container_process_stat",
        )
        async with self._lock:
            pid_map: dict[PID, ContainerId] = {}
            async with aiodocker.Docker() as docker:
                for cid in container_ids:
                    active_pids = await self._get_processes(cid, docker)
                    if cid in self.process_metrics:
                        unused_pids = set(self.process_metrics[cid].keys()) - set(active_pids)
                        if unused_pids:
                            log.debug(
                                "removing pid_metric for {}: {}",
                                cid,
                                ", ".join([str(p) for p in unused_pids]),
                            )
                            self.process_metrics[cid] = {
                                pid_: metric
                                for pid_, metric in self.process_metrics[cid].items()
                                if pid_ in active_pids
                            }
                    for pid_ in active_pids:
                        pid_map[pid_] = cid
            # Here we use asyncio.gather() instead of aiotools.TaskGroup
            # to keep methods of other plugins running when a plugin raises an error
            # instead of cancelling them.
            _tasks: list[asyncio.Task[Sequence[ProcessMeasurement]]] = []
            for computer in self.agent.computers.values():
                _tasks.append(
                    asyncio.create_task(
                        computer.instance.gather_process_measures(
                            self, cast(Mapping[int, str], pid_map)
                        ),
                    )
                )
            self._stage_observer.observe_stage(
                stage="before_gather_measures",
                upper_layer="collect_per_container_process_stat",
            )
            results = await asyncio.gather(*_tasks, return_exceptions=True)
            self._stage_observer.observe_stage(
                stage="before_observe",
                upper_layer="collect_per_container_process_stat",
            )
            updated_cids: Set[ContainerId] = set()
            for result in results:
                if isinstance(result, BaseException):
                    log.error(
                        "collect_per_container_process_stat(): gather_process_measures() error",
                        exc_info=result,
                    )
                    continue
                for proc_measure in result:
                    metric_key = proc_measure.key
                    # update per-process metric
                    for pid, measure in proc_measure.per_process.items():
                        pid = PID(pid)
                        cid = pid_map[pid]
                        updated_cids.add(cid)
                        if cid not in self.process_metrics:
                            self.process_metrics[cid] = {}
                        if pid not in self.process_metrics[cid]:
                            self.process_metrics[cid][pid] = {}
                        if metric_key not in self.process_metrics[cid][pid]:
                            self.process_metrics[cid][pid][metric_key] = Metric(
                                metric_key,
                                proc_measure.type,
                                current=measure.value,
                                capacity=measure.capacity or measure.value,
                                unit_hint=proc_measure.unit_hint,
                                stats=MovingStatistics(measure.value),
                                stats_filter=frozenset(proc_measure.stats_filter),
                                current_hook=proc_measure.current_hook,
                            )
                        else:
                            self.process_metrics[cid][pid][metric_key].update(measure)

            self._stage_observer.observe_stage(
                stage="before_report_to_redis",
                upper_layer="collect_per_container_process_stat",
            )

            # Use ValkeyStatClient set_multiple_keys for batch operations
            key_value_map: dict[str, bytes] = {}
            for cid in updated_cids:
                serializable_table = {}
                for pid in self.process_metrics[cid].keys():
                    metrics = self.process_metrics[cid][pid]
                    serializable_metrics = {}
                    for key, obj in metrics.items():
                        try:
                            serializable_metrics[str(key)] = obj.to_serializable_dict()
                        except ValueError:
                            log.warning("Failed to serialize metric {}: {}", key, str(obj.stats))
                            continue
                    serializable_table[pid] = serializable_metrics
                if self.agent.local_config.debug.log_stats:
                    log.debug(
                        "stats: process_updates: \ncontainer_id: {}\n{}",
                        cid,
                        serializable_table,
                    )
                serialized_metrics = msgpack.packb(serializable_table)
                key_value_map[str(cid)] = serialized_metrics

            if key_value_map:
                await self.agent.valkey_stat_client.set_multiple_keys(key_value_map, expire_sec=8)



"""
Config files stage for kernel lifecycle.

This stage handles creation of environment and resource configuration files.
"""

import asyncio
import shutil
from dataclasses import dataclass
from io import StringIO
from pathlib import Path
from typing import Mapping, override

from ai.backend.agent.resources import ComputerContext, KernelResourceSpec
from ai.backend.common.stage.types import ArgsSpecGenerator, Provisioner, ProvisionStage
from ai.backend.common.types import DeviceName


@dataclass
class ConfigFileSpec:
    config_dir: Path
    environ: Mapping[str, str]
    resource_spec: KernelResourceSpec
    computers: Mapping[DeviceName, ComputerContext]
    accelerator_envs: Mapping[str, str]  # Additional envs from accelerators


class ConfigFileSpecGenerator(ArgsSpecGenerator[ConfigFileSpec]):
    pass


@dataclass
class ConfigFileResult:
    environ_path: Path
    resource_path: Path
    environ_base_path: Path
    resource_base_path: Path


class ConfigFileProvisioner(Provisioner[ConfigFileSpec, ConfigFileResult]):
    """
    Provisioner for configuration file creation.

    Creates environ.txt and resource.txt files with environment variables
    and resource specifications.
    """

    @property
    @override
    def name(self) -> str:
        return "docker-config-files"

    @override
    async def setup(self, spec: ConfigFileSpec) -> ConfigFileResult:
        await self._write_environ(spec)
        await self._write_resource_txt(spec)
        await self._write_backup_files(spec)

        return ConfigFileResult(
            environ_path=self._environ_path(spec),
            resource_path=self._resource_path(spec),
            environ_base_path=self._environ_base_path(spec),
            resource_base_path=self._resource_base_path(spec),
        )

    def _environ_path(self, spec: ConfigFileSpec) -> Path:
        return spec.config_dir / "environ.txt"

    def _environ_base_path(self, spec: ConfigFileSpec) -> Path:
        return spec.config_dir / "environ_base.txt"

    def _resource_path(self, spec: ConfigFileSpec) -> Path:
        return spec.config_dir / "resource.txt"

    def _resource_base_path(self, spec: ConfigFileSpec) -> Path:
        return spec.config_dir / "resource_base.txt"

    async def _write_environ(self, spec: ConfigFileSpec) -> None:
        loop = asyncio.get_running_loop()
        with StringIO() as buf:
            # Write basic environment variables
            for k, v in spec.environ.items():
                buf.write(f"{k}={v}\n")

            # Write accelerator environment variables
            for k, v in spec.accelerator_envs.items():
                buf.write(f"{k}={v}\n")

            await loop.run_in_executor(
                None,
                (self._environ_path(spec)).write_bytes,
                buf.getvalue().encode("utf8"),
            )

    async def _write_resource_txt(self, spec: ConfigFileSpec) -> None:
        loop = asyncio.get_running_loop()
        with StringIO() as buf:
            spec.resource_spec.write_to_file(buf)

            # Add device-specific resource data
            for dev_type, device_alloc in spec.resource_spec.allocations.items():
                if dev_type in spec.computers:
                    device_plugin = spec.computers[dev_type].instance
                    kvpairs = await device_plugin.generate_resource_data(device_alloc)
                    for k, v in kvpairs.items():
                        buf.write(f"{k}={v}\n")

            await loop.run_in_executor(
                None,
                (self._resource_path(spec)).write_bytes,
                buf.getvalue().encode("utf8"),
            )

    async def _write_backup_files(self, spec: ConfigFileSpec) -> None:
        # Create backup files for environ and resource
        environ_path = self._environ_path(spec)
        environ_base_path = self._environ_base_path(spec)
        resource_path = self._resource_path(spec)
        resource_base_path = self._resource_base_path(spec)

        loop = asyncio.get_running_loop()
        await loop.run_in_executor(None, shutil.copyfile, environ_path, environ_base_path)
        await loop.run_in_executor(None, shutil.copyfile, resource_path, resource_base_path)

    @override
    async def teardown(self, resource: ConfigFileResult) -> None:
        # Config files are cleaned up with scratch directory
        pass


class ConfigFileStage(ProvisionStage[ConfigFileSpec, ConfigFileResult]):
    """
    Stage for creating configuration files in kernel containers.
    """

    pass



from ai.backend.common.plugin.monitor import ErrorPluginContext, StatsPluginContext


class AgentErrorPluginContext(ErrorPluginContext):
    blocklist = {"ai.backend.manager"}


class AgentStatsPluginContext(StatsPluginContext):
    blocklist = {"ai.backend.manager"}



import asyncio
import functools
import os
import shutil
import subprocess
import sys
from collections.abc import Iterable
from dataclasses import dataclass
from pathlib import Path
from subprocess import CalledProcessError
from typing import Optional, override

import pkg_resources

from ai.backend.common.docker import KernelFeatures
from ai.backend.common.stage.types import (
    ArgsSpecGenerator,
    Provisioner,
    ProvisionStage,
)
from ai.backend.common.types import BinarySize, KernelId

from .utils import ScratchUtil


@dataclass
class ContainerOwnershipConfig:
    kernel_uid: Optional[int]
    kernel_gid: Optional[int]
    supplementary_gids: set[int]

    fallback_kernel_uid: int
    fallback_kernel_gid: int

    kernel_features: frozenset[str]


@dataclass
class ScratchSpec:
    kernel_id: KernelId

    container_config: ContainerOwnershipConfig

    scratch_type: str
    scratch_root: Path
    scratch_size: BinarySize


class ScratchSpecGenerator(ArgsSpecGenerator[ScratchSpec]):
    pass


@dataclass
class ScratchResult:
    scratch_dir: Path
    scratch_file: Path
    tmp_dir: Path
    work_dir: Path
    config_dir: Path

    scratch_type: str


class ScratchProvisioner(Provisioner[ScratchSpec, ScratchResult]):
    """
    Provisioner for the kernel creation setup stage.
    This is a no-op provisioner as it does not create any resources.
    """

    @property
    @override
    def name(self) -> str:
        return "docker-scratch"

    @override
    async def setup(self, spec: ScratchSpec) -> ScratchResult:
        await self._create_filesystem(spec)
        await self._create_scratch_dirs(spec)
        await self._clone_dotfiles(spec)
        return ScratchResult(
            scratch_dir=ScratchUtil.scratch_dir(spec.scratch_root, spec.kernel_id),
            scratch_file=ScratchUtil.scratch_file(spec.scratch_root, spec.kernel_id),
            tmp_dir=ScratchUtil.tmp_dir(spec.scratch_root, spec.kernel_id),
            work_dir=ScratchUtil.work_dir(spec.scratch_root, spec.kernel_id),
            config_dir=ScratchUtil.config_dir(spec.scratch_root, spec.kernel_id),
            scratch_type=spec.scratch_type,
        )

    async def _create_filesystem(self, spec: ScratchSpec) -> None:
        loop = asyncio.get_running_loop()
        tmp_dir = ScratchUtil.tmp_dir(spec.scratch_root, spec.kernel_id)
        scratch_dir = ScratchUtil.scratch_dir(spec.scratch_root, spec.kernel_id)
        if sys.platform.startswith("linux") and spec.scratch_type == "memory":
            await loop.run_in_executor(None, functools.partial(tmp_dir.mkdir, exist_ok=True))
            await self._create_scratch_filesystem(scratch_dir, 64)
            await self._create_scratch_filesystem(tmp_dir, 64)
        elif sys.platform.startswith("linux") and spec.scratch_type == "hostfile":
            await self._create_loop_filesystem(spec.scratch_root, spec.scratch_size, spec.kernel_id)
        else:
            await loop.run_in_executor(None, functools.partial(scratch_dir.mkdir, exist_ok=True))

    async def _create_scratch_filesystem(self, scratch_dir: Path, size: int) -> None:
        """
        Create scratch folder size quota by using tmpfs filesystem.

        :param scratch_dir: The path of scratch directory.

        :param size: The quota size of scratch directory.
                    Size parameter is must be MiB(mebibyte).
        """

        proc = await asyncio.create_subprocess_exec(*[
            "mount",
            "-t",
            "tmpfs",
            "-o",
            f"size={size}M",
            "tmpfs",
            f"{scratch_dir}",
        ])
        exit_code = await proc.wait()

        if exit_code < 0:
            raise CalledProcessError(
                exit_code,
                proc.args,  # type: ignore[attr-defined]
                output=proc.stdout,  # type: ignore[arg-type]
                stderr=proc.stderr,  # type: ignore[arg-type]
            )

    def _create_sparse_file(self, name: str, size: int) -> None:
        fd = os.open(name, os.O_CREAT, 0o644)
        os.close(fd)
        os.truncate(name, size)
        # Check that no space was allocated
        stat = os.stat(name)
        if stat.st_blocks != 0:
            raise RuntimeError("could not create sparse file")

    async def _create_loop_filesystem(
        self, scratch_root: Path, scratch_size: int, kernel_id: KernelId
    ) -> None:
        loop = asyncio.get_running_loop()
        scratch_dir = ScratchUtil.scratch_dir(scratch_root, kernel_id)
        scratch_file = ScratchUtil.scratch_file(scratch_root, kernel_id)
        await loop.run_in_executor(
            None, functools.partial(os.makedirs, str(scratch_dir), exist_ok=True)
        )
        await loop.run_in_executor(None, self._create_sparse_file, str(scratch_file), scratch_size)
        mkfs = await asyncio.create_subprocess_exec(
            "/sbin/mkfs.ext4",
            str(scratch_file),
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        )
        exit_code = await mkfs.wait()
        if exit_code != 0:
            raise RuntimeError("mkfs failed")
        mount = await asyncio.create_subprocess_exec("mount", str(scratch_file), str(scratch_dir))
        exit_code = await mount.wait()
        if exit_code != 0:
            raise RuntimeError("mount failed")

    async def _create_scratch_dirs(self, spec: ScratchSpec) -> None:
        work_dir = ScratchUtil.work_dir(spec.scratch_root, spec.kernel_id)
        config_dir = ScratchUtil.config_dir(spec.scratch_root, spec.kernel_id)

        def create() -> None:
            config_dir.mkdir(parents=True, exist_ok=True)
            config_dir.chmod(0o755)
            work_dir.mkdir(parents=True, exist_ok=True)
            work_dir.chmod(0o755)

        loop = asyncio.get_running_loop()
        await loop.run_in_executor(None, create)

    async def _clone_dotfiles(self, spec: ScratchSpec) -> None:
        # Since these files are bind-mounted inside a bind-mounted directory,
        # we need to touch them first to avoid their "ghost" files are created
        # as root in the host-side filesystem, which prevents deletion of scratch
        # directories when the agent is running as non-root.

        loop = asyncio.get_running_loop()
        await loop.run_in_executor(None, self._clone_func, spec)

    def _clone_func(self, spec: ScratchSpec) -> None:
        work_dir = ScratchUtil.work_dir(spec.scratch_root, spec.kernel_id)
        jupyter_custom_css_path = Path(
            pkg_resources.resource_filename("ai.backend.runner", "jupyter-custom.css")
        )
        logo_path = Path(pkg_resources.resource_filename("ai.backend.runner", "logo.svg"))
        font_path = Path(pkg_resources.resource_filename("ai.backend.runner", "roboto.ttf"))
        font_italic_path = Path(
            pkg_resources.resource_filename("ai.backend.runner", "roboto-italic.ttf")
        )
        bashrc_path = Path(pkg_resources.resource_filename("ai.backend.runner", ".bashrc"))
        bash_profile_path = Path(
            pkg_resources.resource_filename("ai.backend.runner", ".bash_profile")
        )
        zshrc_path = Path(pkg_resources.resource_filename("ai.backend.runner", ".zshrc"))
        vimrc_path = Path(pkg_resources.resource_filename("ai.backend.runner", ".vimrc"))
        tmux_conf_path = Path(pkg_resources.resource_filename("ai.backend.runner", ".tmux.conf"))
        jupyter_custom_dir = work_dir / ".jupyter" / "custom"
        jupyter_custom_dir.mkdir(parents=True, exist_ok=True)
        shutil.copy(jupyter_custom_css_path.resolve(), jupyter_custom_dir / "custom.css")
        shutil.copy(logo_path.resolve(), jupyter_custom_dir / "logo.svg")
        shutil.copy(font_path.resolve(), jupyter_custom_dir / "roboto.ttf")
        shutil.copy(font_italic_path.resolve(), jupyter_custom_dir / "roboto-italic.ttf")
        shutil.copy(bashrc_path.resolve(), work_dir / ".bashrc")
        shutil.copy(bash_profile_path.resolve(), work_dir / ".bash_profile")
        shutil.copy(zshrc_path.resolve(), work_dir / ".zshrc")
        shutil.copy(vimrc_path.resolve(), work_dir / ".vimrc")
        shutil.copy(tmux_conf_path.resolve(), work_dir / ".tmux.conf")

        paths = [
            work_dir,
            work_dir / ".jupyter",
            work_dir / ".jupyter" / "custom",
            work_dir / ".bashrc",
            work_dir / ".bash_profile",
            work_dir / ".zshrc",
            work_dir / ".vimrc",
            work_dir / ".tmux.conf",
        ]
        self._chown_paths_if_root(paths, spec.container_config)

    def _chown_paths_if_root(
        self,
        paths: Iterable[Path],
        config: ContainerOwnershipConfig,
    ) -> None:
        valid_uid: Optional[int]
        valid_gid: Optional[int]
        if os.geteuid() == 0:  # only possible when I am root.
            if KernelFeatures.UID_MATCH in config.kernel_features:
                valid_uid = (
                    config.kernel_uid
                    if config.kernel_uid is not None
                    else config.fallback_kernel_uid
                )
                valid_gid = (
                    config.kernel_gid
                    if config.kernel_gid is not None
                    else config.fallback_kernel_gid
                )
            else:
                valid_uid = config.kernel_uid
                valid_gid = config.kernel_gid
            for p in paths:
                if valid_uid is None or valid_gid is None:
                    stat = os.stat(p)
                    valid_uid = stat.st_uid if valid_uid is None else valid_uid
                    valid_gid = stat.st_gid if valid_gid is None else valid_gid
                os.chown(p, valid_uid, valid_gid)

    @override
    async def teardown(self, resource: ScratchResult) -> None:
        loop = asyncio.get_running_loop()
        scratch_dir = resource.scratch_dir
        tmp_dir = resource.tmp_dir
        try:
            if sys.platform.startswith("linux") and resource.scratch_type == "memory":
                await self._destroy_scratch_filesystem(scratch_dir)
                await self._destroy_scratch_filesystem(tmp_dir)
                await loop.run_in_executor(None, shutil.rmtree, scratch_dir)
                await loop.run_in_executor(None, shutil.rmtree, tmp_dir)
            elif sys.platform.startswith("linux") and resource.scratch_type == "hostfile":
                await self._destroy_loop_filesystem(resource)
            else:
                await loop.run_in_executor(None, shutil.rmtree, scratch_dir)
        except CalledProcessError:
            pass
        except FileNotFoundError:
            pass

    async def _destroy_scratch_filesystem(self, scratch_dir: Path) -> None:
        """
        Destroy scratch folder size quota by using tmpfs filesystem.

        :param scratch_dir: The path of scratch directory.
        """
        proc = await asyncio.create_subprocess_exec(*[
            "umount",
            f"{scratch_dir}",
        ])
        exit_code = await proc.wait()

        if exit_code < 0:
            raise CalledProcessError(
                exit_code,
                proc.args,  # type: ignore[attr-defined]
                output=proc.stdout,  # type: ignore[arg-type]
                stderr=proc.stderr,  # type: ignore[arg-type]
            )

    async def _destroy_loop_filesystem(self, resource: ScratchResult) -> None:
        loop = asyncio.get_running_loop()
        scratch_dir = resource.scratch_dir
        scratch_file = resource.scratch_file
        umount = await asyncio.create_subprocess_exec("umount", str(scratch_dir))
        exit_code = await umount.wait()
        if exit_code != 0:
            raise RuntimeError("umount failed")
        await loop.run_in_executor(None, os.remove, str(scratch_file))
        await loop.run_in_executor(None, shutil.rmtree, str(scratch_dir))


class ScratchStage(ProvisionStage[ScratchSpec, ScratchResult]):
    pass



import asyncio
import enum
import functools
import uuid
from collections.abc import Iterable
from typing import Optional, Self

from prometheus_client import Counter, Gauge, Histogram

from ai.backend.common.metrics.types import (
    CONTAINER_UTILIZATION_METRIC_LABEL_NAME,
    DEVICE_UTILIZATION_METRIC_LABEL_NAME,
    UNDEFINED,
    UTILIZATION_METRIC_DETENTION,
)
from ai.backend.common.types import AgentId, KernelId, MetricKey, SessionId

from .types import (
    ALL_METRIC_VALUE_TYPES,
    FlattenedDeviceMetric,
    FlattenedKernelMetric,
)


class StatScope(enum.StrEnum):
    NODE = "node"
    CONTAINER = "container"
    PROCESS = "process"


class RPCMetricObserver:
    _instance: Optional[Self] = None

    _rpc_requests: Counter
    _rpc_failure_requests: Counter
    _rpc_request_duration: Histogram

    def __init__(self) -> None:
        self._rpc_requests = Counter(
            name="backendai_rpc_requests_total",
            documentation="Number of RPC requests",
            labelnames=["method"],
        )
        self._rpc_failure_requests = Counter(
            name="backendai_rpc_failure_requests_total",
            documentation="Number of failed RPC requests",
            labelnames=["method", "exception"],
        )
        self._rpc_request_duration = Histogram(
            name="backendai_rpc_request_duration_seconds",
            documentation="Duration of RPC requests",
            labelnames=["method"],
            buckets=[0.1, 1, 10, 30, 60, 300, 600],
        )

    @classmethod
    def instance(cls) -> Self:
        if cls._instance is None:
            cls._instance = cls()
        return cls._instance

    def observe_rpc_request_success(self, *, method: str, duration: float) -> None:
        self._rpc_requests.labels(method=method).inc()
        self._rpc_request_duration.labels(method=method).observe(duration)

    def observe_rpc_request_failure(
        self, *, method: str, duration: float, exception: BaseException
    ) -> None:
        exception_name = exception.__class__.__name__
        self._rpc_requests.labels(method=method).inc()
        self._rpc_failure_requests.labels(method=method, exception=exception_name).inc()
        self._rpc_request_duration.labels(method=method).observe(duration)


class UtilizationMetricObserver:
    _instance: Optional[Self] = None
    _removal_tasks: dict[KernelId, asyncio.Task[None]]

    _container_metric: Gauge
    _device_metric: Gauge

    def __init__(self) -> None:
        self._removal_tasks = {}
        self._container_metric = Gauge(
            name="backendai_container_utilization",
            documentation="Container utilization metrics",
            labelnames=[
                CONTAINER_UTILIZATION_METRIC_LABEL_NAME,
                "agent_id",
                "kernel_id",
                "session_id",
                "user_id",
                "project_id",
                "value_type",
            ],
        )
        self._device_metric = Gauge(
            name="backendai_device_utilization",
            documentation="Device utilization metrics",
            labelnames=[
                DEVICE_UTILIZATION_METRIC_LABEL_NAME,
                "agent_id",
                "device_id",
                "value_type",
            ],
        )

    @classmethod
    def instance(cls) -> Self:
        if cls._instance is None:
            cls._instance = cls()
        return cls._instance

    def observe_container_metric(
        self,
        *,
        metric: FlattenedKernelMetric,
    ) -> None:
        for metric_value_type, value in metric.value_pairs:
            self._container_metric.labels(
                container_metric_name=metric.key,
                agent_id=metric.agent_id,
                kernel_id=metric.kernel_id,
                session_id=metric.session_id or UNDEFINED,
                user_id=metric.owner_user_id or UNDEFINED,
                project_id=metric.project_id or UNDEFINED,
                value_type=metric_value_type,
            ).set(float(value))

    async def lazy_remove_container_metric(
        self,
        *,
        agent_id: AgentId,
        kernel_id: KernelId,
        session_id: Optional[SessionId],
        owner_user_id: Optional[uuid.UUID],
        project_id: Optional[uuid.UUID],
        keys: Iterable[MetricKey],
    ) -> None:
        async def remove_later() -> None:
            await asyncio.sleep(UTILIZATION_METRIC_DETENTION)
            for key in keys:
                for value_type in ALL_METRIC_VALUE_TYPES:
                    try:
                        self._container_metric.remove(
                            key,
                            agent_id,
                            kernel_id,
                            session_id or UNDEFINED,
                            owner_user_id or UNDEFINED,
                            project_id or UNDEFINED,
                            value_type,
                        )
                    except KeyError:
                        continue

        def callback(task: asyncio.Task) -> None:
            self._removal_tasks.pop(kernel_id, None)

        task = asyncio.create_task(remove_later())
        self._removal_tasks[kernel_id] = task
        task.add_done_callback(functools.partial(callback))

    def observe_device_metric(
        self,
        *,
        metric: FlattenedDeviceMetric,
    ) -> None:
        for metric_value_type, value in metric.value_pairs:
            self._device_metric.labels(
                device_metric_name=metric.key,
                agent_id=metric.agent_id,
                device_id=metric.device_id,
                value_type=metric_value_type,
            ).set(float(value))


class SyncContainerLifecycleObserver:
    _instance: Optional[Self] = None

    _task_trigger_count: Counter
    _task_success_count: Counter
    _task_failure_count: Counter

    def __init__(self) -> None:
        self._task_trigger_count = Counter(
            name="backendai_sync_container_lifecycle_trigger_count",
            documentation="Number of sync_container_lifecycle() task triggered",
            labelnames=["agent_id"],
        )
        self._task_success_count = Counter(
            name="backendai_sync_container_lifecycle_success_count",
            documentation="Number of sync_container_lifecycle() task succeeded",
            labelnames=["agent_id"],
        )
        self._task_failure_count = Counter(
            name="backendai_sync_container_lifecycle_failure_count",
            documentation="Number of sync_container_lifecycle() task failed",
            labelnames=["agent_id", "exception"],
        )

    @classmethod
    def instance(cls) -> Self:
        if cls._instance is None:
            cls._instance = cls()
        return cls._instance

    def observe_container_lifecycle_triggered(
        self,
        *,
        agent_id: AgentId,
    ) -> None:
        self._task_trigger_count.labels(agent_id=agent_id).inc()

    def observe_container_lifecycle_success(
        self,
        *,
        agent_id: AgentId,
        num_synced_kernels: int,
    ) -> None:
        self._task_success_count.labels(agent_id=agent_id).inc(amount=num_synced_kernels)

    def observe_container_lifecycle_failure(
        self,
        *,
        agent_id: AgentId,
        exception: BaseException,
    ) -> None:
        exception_name = exception.__class__.__name__
        self._task_failure_count.labels(agent_id=agent_id, exception=exception_name).inc()


class StatTaskObserver:
    _instance: Optional[Self] = None

    _task_trigger_count: Counter
    _task_success_count: Counter
    _task_failure_count: Counter

    def __init__(self) -> None:
        self._task_trigger_count = Counter(
            name="backendai_stat_task_trigger_count",
            documentation="Number of stat() task triggered",
            labelnames=["agent_id", "stat_scope"],
        )
        self._task_success_count = Counter(
            name="backendai_stat_task_success_count",
            documentation="Number of stat() task succeeded",
            labelnames=["agent_id", "stat_scope"],
        )
        self._task_failure_count = Counter(
            name="backendai_stat_task_failure_count",
            documentation="Number of stat() task failed",
            labelnames=["agent_id", "stat_scope", "exception"],
        )

    @classmethod
    def instance(cls) -> Self:
        if cls._instance is None:
            cls._instance = cls()
        return cls._instance

    def observe_stat_task_triggered(
        self,
        *,
        agent_id: AgentId,
        stat_scope: StatScope,
    ) -> None:
        self._task_trigger_count.labels(agent_id=agent_id, stat_scope=stat_scope).inc()

    def observe_stat_task_success(
        self,
        *,
        agent_id: AgentId,
        stat_scope: StatScope,
    ) -> None:
        self._task_success_count.labels(agent_id=agent_id, stat_scope=stat_scope).inc()

    def observe_stat_task_failure(
        self,
        *,
        agent_id: AgentId,
        stat_scope: StatScope,
        exception: BaseException,
    ) -> None:
        exception_name = exception.__class__.__name__
        self._task_failure_count.labels(
            agent_id=agent_id, stat_scope=stat_scope, exception=exception_name
        ).inc()



from dataclasses import dataclass
from typing import Optional, override

from ai.backend.agent.config.unified import ContainerSandboxType
from ai.backend.common.stage.types import (
    ArgsSpecGenerator,
    Provisioner,
    ProvisionStage,
)


@dataclass
class CmdArgSpec:
    runtime_type: str
    runtime_path: Optional[str]
    sandbox_type: ContainerSandboxType
    jail_args: list[str]
    debug_kernel_runner: bool


class CmdArgSpecGenerator(ArgsSpecGenerator[CmdArgSpec]):
    pass


@dataclass
class CmdArgResult:
    cmdargs: list[str]


class CmdArgProvisioner(Provisioner[CmdArgSpec, CmdArgResult]):
    @property
    @override
    def name(self) -> str:
        return "docker-cmdarg"

    @override
    async def setup(self, spec: CmdArgSpec) -> CmdArgResult:
        cmdargs: list[str] = []
        krunner_opts: list[str] = []
        if spec.sandbox_type == ContainerSandboxType.JAIL:
            cmdargs += [
                "/opt/kernel/jail",
                # "--policy",
                # "/etc/backend.ai/jail/policy.yml",
                # TODO: Update default Jail policy in images
            ]
            if spec.jail_args:
                cmdargs += map(lambda s: s.strip(), spec.jail_args)
            cmdargs += ["--"]
        if spec.debug_kernel_runner:
            krunner_opts.append("--debug")
        cmdargs += [
            "/opt/backend.ai/bin/python",
            "-s",
            "-m",
            "ai.backend.kernel",
            *krunner_opts,
            spec.runtime_type,
        ]
        if spec.runtime_path is not None:
            cmdargs.append(spec.runtime_path)
        return CmdArgResult(cmdargs=cmdargs)

    @override
    async def teardown(self, resource: CmdArgResult) -> None:
        pass


class CmdArgStage(ProvisionStage[CmdArgSpec, CmdArgResult]):
    """
    Stage for generating command-line arguments for the kernel runner.
    This stage prepares the command arguments based on the provided specifications.
    """



import asyncio
from subprocess import CalledProcessError


async def create_scratch_filesystem(scratch_dir, size):
    """
    Create scratch folder size quota by using tmpfs filesystem.

    :param scratch_dir: The path of scratch directory.

    :param size: The quota size of scratch directory.
                 Size parameter is must be MiB(mebibyte).
    """

    proc = await asyncio.create_subprocess_exec(*[
        "mount",
        "-t",
        "tmpfs",
        "-o",
        f"size={size}M",
        "tmpfs",
        f"{scratch_dir}",
    ])
    exit_code = await proc.wait()

    if exit_code < 0:
        raise CalledProcessError(proc.returncode, proc.args, output=proc.stdout, stderr=proc.stderr)


async def destroy_scratch_filesystem(scratch_dir):
    """
    Destroy scratch folder size quota by using tmpfs filesystem.

    :param scratch_dir: The path of scratch directory.
    """
    proc = await asyncio.create_subprocess_exec(*[
        "umount",
        f"{scratch_dir}",
    ])
    exit_code = await proc.wait()

    if exit_code < 0:
        raise CalledProcessError(proc.returncode, proc.args, output=proc.stdout, stderr=proc.stderr)



"""
Credentials stage for kernel lifecycle.

This stage handles Docker credentials setup for containers.
"""

import asyncio
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Mapping, Optional, override

from ai.backend.common.json import dump_json
from ai.backend.common.stage.types import ArgsSpecGenerator, Provisioner, ProvisionStage


@dataclass
class CredentialsSpec:
    config_dir: Path
    docker_credentials: Optional[Mapping[str, Any]]


class CredentialsSpecGenerator(ArgsSpecGenerator[CredentialsSpec]):
    pass


@dataclass
class CredentialsResult:
    credentials_path: Optional[Path]


class CredentialsProvisioner(Provisioner[CredentialsSpec, CredentialsResult]):
    """
    Provisioner for Docker credentials management.

    Writes Docker credentials to config directory if provided.
    """

    @property
    @override
    def name(self) -> str:
        return "docker-credentials"

    @override
    async def setup(self, spec: CredentialsSpec) -> CredentialsResult:
        if not spec.docker_credentials:
            return CredentialsResult(credentials_path=None)

        loop = asyncio.get_running_loop()
        credentials_path = spec.config_dir / "docker-creds.json"

        await loop.run_in_executor(
            None,
            credentials_path.write_bytes,
            dump_json(spec.docker_credentials),
        )

        return CredentialsResult(credentials_path=credentials_path)

    @override
    async def teardown(self, resource: CredentialsResult) -> None:
        # Credentials file is cleaned up with scratch directory
        pass


class CredentialsStage(ProvisionStage[CredentialsSpec, CredentialsResult]):
    """
    Stage for managing Docker credentials in kernel containers.
    """

    pass



from collections.abc import Collection, Mapping, Sequence
from dataclasses import dataclass
from decimal import Decimal
from typing import Any, Final, Optional, Self, override

from ai.backend.agent.resources import ComputerContext, KernelResourceSpec
from ai.backend.common.docker import KernelFeatures, LabelName
from ai.backend.common.stage.types import (
    ArgsSpecGenerator,
    Provisioner,
    ProvisionStage,
)
from ai.backend.common.types import (
    BinarySize,
    DeviceModelInfo,
    DeviceName,
    KernelCreationConfig,
    SlotName,
)

from ..defs import LIBBAIHOOK_MOUNT_PATH

LD_PRELOAD: Final[str] = "LD_PRELOAD"
LOCAL_USER_ID: Final[str] = "LOCAL_USER_ID"
LOCAL_GROUP_ID: Final[str] = "LOCAL_GROUP_ID"
ADDITIONAL_GIDS: Final[str] = "ADDITIONAL_GIDS"


@dataclass
class AgentInfo:
    computers: Mapping[DeviceName, ComputerContext]
    distro: str
    architecture: str
    kernel_uid: int
    kernel_gid: int


@dataclass
class KernelInfo:
    kernel_creation_config: KernelCreationConfig
    kernel_features: frozenset[str]
    resource_spec: KernelResourceSpec
    overriding_uid: Optional[int]
    overriding_gid: Optional[int]
    supplementary_gids: set[int]


@dataclass
class EnvironSpec:
    agent_info: AgentInfo
    kernel_info: KernelInfo


class EnvironSpecGenerator(ArgsSpecGenerator[EnvironSpec]):
    pass


@dataclass
class EnvironResult:
    environ: dict[str, str]


class Environ(dict[str, str]):
    def set_value(self, key: str, value: Optional[Any]) -> Self:
        if value is None:
            return self
        self[key] = str(value)
        return self

    def append_values(self, key: str, values: Collection[str], *, separator: str) -> Self:
        if not values:
            return self
        if orig_value := self.get(key):
            # If the key already exists, append the new values
            orig_val = {v for v in orig_value.split(separator) if v}
            new_values = orig_val | set(values)
        else:
            # If the key does not exist, create a new entry
            new_values = set(values)
        self[key] = separator.join(sorted(new_values))
        return self

    def update_always(self, environ: Mapping[str, str]) -> Self:
        for key, value in environ.items():
            self[key] = value
        return self

    def update_if_not_exists(self, environ: Mapping[str, str]) -> Self:
        for key, value in environ.items():
            if key not in self:
                self[key] = value
        return self

    def to_dict(self) -> dict[str, str]:
        return dict(self)


class EnvironProvisioner(Provisioner[EnvironSpec, EnvironResult]):
    """
    Provisioner for the kernel creation setup stage.
    This is a no-op provisioner as it does not create any resources.
    """

    @property
    @override
    def name(self) -> str:
        return "environ"

    @override
    async def setup(self, spec: EnvironSpec) -> EnvironResult:
        environ = Environ({
            **spec.kernel_info.kernel_creation_config["environ"]
        })  # Start with the base environment
        environ = (
            environ.set_value(LD_PRELOAD, LIBBAIHOOK_MOUNT_PATH)
            .set_value(LOCAL_USER_ID, self._get_local_uid(spec))
            .set_value(LOCAL_GROUP_ID, self._get_local_gid(spec))
            .append_values(ADDITIONAL_GIDS, self._get_supplementary_gids(spec), separator=",")
            .append_values(ADDITIONAL_GIDS, self._get_computer_gids(spec), separator=",")
            .update_if_not_exists(self._get_core_count(spec))
        )

        hook_paths = await self._get_container_hooks(spec)
        device_environ = await self._get_device_environ(spec)
        environ = environ.append_values(LD_PRELOAD, hook_paths, separator=":").update_always(
            device_environ
        )
        return EnvironResult(environ=environ.to_dict())

    def _get_local_uid(self, spec: EnvironSpec) -> Optional[int]:
        if spec.kernel_info.overriding_uid is not None:
            return spec.kernel_info.overriding_uid
        if KernelFeatures.UID_MATCH in spec.kernel_info.kernel_features:
            return spec.agent_info.kernel_uid
        return None

    def _get_local_gid(self, spec: EnvironSpec) -> Optional[int]:
        if spec.kernel_info.overriding_gid is not None:
            return spec.kernel_info.overriding_gid
        if KernelFeatures.UID_MATCH in spec.kernel_info.kernel_features:
            return spec.agent_info.kernel_gid
        return None

    def _get_supplementary_gids(self, spec: EnvironSpec) -> set[str]:
        return {str(gid) for gid in spec.kernel_info.supplementary_gids}

    def _get_computer_gids(self, spec: EnvironSpec) -> set[str]:
        additional_gid_set: set[int] = set()
        for dev_type in spec.kernel_info.resource_spec.allocations:
            computer_ctx = spec.agent_info.computers[dev_type]
            additional_gids = computer_ctx.instance.get_additional_gids()
            additional_gid_set.update(additional_gids)
        return {str(gid) for gid in additional_gid_set}

    def _get_core_count(self, spec: EnvironSpec) -> dict[str, str]:
        image_labels = spec.kernel_info.kernel_creation_config["image"]["labels"]
        label_envs_corecount = image_labels.get(LabelName.ENVS_CORECOUNT, "")
        envs_corecount = label_envs_corecount.split(",") if label_envs_corecount else []
        cpu_core_count = len(
            spec.kernel_info.resource_spec.allocations[DeviceName("cpu")][SlotName("cpu")]
        )
        return {k: str(cpu_core_count) for k in envs_corecount}

    async def _get_container_hooks(self, spec: EnvironSpec) -> set[str]:
        container_hook_path_set: set[str] = set()
        for dev_type, device_alloc in spec.kernel_info.resource_spec.allocations.items():
            alloc_sum = Decimal(0)
            for per_dev_alloc in device_alloc.values():
                alloc_sum += sum(per_dev_alloc.values())
            do_hook_mount = alloc_sum > 0
            if not do_hook_mount:
                continue
            computer_ctx = spec.agent_info.computers[dev_type]
            hook_paths = await computer_ctx.instance.get_hooks(
                spec.agent_info.distro, spec.agent_info.architecture
            )
            for p in hook_paths:
                hook_path = p.absolute()
                final_path = f"/opt/kernel/{hook_path.name}"
                if final_path in container_hook_path_set:
                    continue
                container_hook_path_set.add(final_path)
        return container_hook_path_set

    async def _get_device_environ(self, spec: EnvironSpec) -> dict[str, str]:
        environ: dict[str, str] = {}
        # Get attached devices information (including model_name).
        attached_devices: dict[DeviceName, Sequence[DeviceModelInfo]] = {}
        for alloc_dev_name, device_alloc in spec.kernel_info.resource_spec.allocations.items():
            computer_ctx = spec.agent_info.computers[alloc_dev_name]
            devices = await computer_ctx.instance.get_attached_devices(device_alloc)
            attached_devices[alloc_dev_name] = devices

        # Generate GPU config env-vars
        has_gpu_config = False
        for dev_name, attached_accelerators in attached_devices.items():
            if has_gpu_config:
                # Generate GPU config for the first-seen accelerator only
                continue
            if dev_name in (DeviceName("cpu"), DeviceName("mem")):
                # Skip intrinsic slots
                continue
            mem_per_device: list[str] = []
            mem_per_device_tf: list[str] = []
            # proc_items = []  # (unused yet)
            for local_idx, dev_info in enumerate(attached_accelerators):
                mem = BinarySize(dev_info["data"].get("mem", 0))
                mem_per_device.append(f"{local_idx}:{mem:s}")
                mem_in_megibytes = f"{mem // (2**20):d}"
                mem_per_device_tf.append(f"{local_idx}:{mem_in_megibytes}")
                # The processor count is not used yet!
                # NOTE: Keep backward-compatibility with the CUDA plugin ("smp")
                # proc = dev_info["data"].get("proc", dev_info["data"].get("smp", 0))
                # proc_items.append(f"{local_idx}:{proc}")
            if attached_accelerators:
                # proc_str = ",".join(proc_items)  # (unused yet)
                environ["GPU_TYPE"] = str(dev_name)
                environ["GPU_MODEL_NAME"] = attached_accelerators[0]["model_name"]
                environ["GPU_CONFIG"] = ",".join(mem_per_device)
                environ["TF_GPU_MEMORY_ALLOC"] = ",".join(mem_per_device_tf)
            environ["GPU_COUNT"] = str(len(attached_accelerators))
            environ["N_GPUS"] = str(len(attached_accelerators))
            has_gpu_config = True
        if not has_gpu_config:
            environ["GPU_COUNT"] = "0"
            environ["N_GPUS"] = "0"
        return environ

    @override
    async def teardown(self, resource: EnvironResult) -> None:
        pass


class EnvironStage(ProvisionStage[EnvironSpec, EnvironResult]):
    pass



from dataclasses import dataclass
from pathlib import Path
from typing import override

from ai.backend.agent.resources import Mount
from ai.backend.common.stage.types import ArgsSpecGenerator, Provisioner, ProvisionStage
from ai.backend.common.types import (
    MountTypes,
    VFolderMount,
)


@dataclass
class VFolderMountSpec:
    mounts: list[VFolderMount]
    prevent_vfolder_mount: bool


class VFolderMountSpecGenerator(ArgsSpecGenerator[VFolderMountSpec]):
    pass


@dataclass
class VFolderMountResult:
    mounts: list[Mount]


class VFolderMountProvisioner(Provisioner[VFolderMountSpec, VFolderMountResult]):
    """
    Provisioner for the kernel creation setup stage.
    This is a no-op provisioner as it does not create any resources.
    """

    @property
    @override
    def name(self) -> str:
        return "docker-vfolder-mount"

    @override
    async def setup(self, spec: VFolderMountSpec) -> VFolderMountResult:
        result: list[Mount] = []
        for vfolder in spec.mounts:
            if spec.prevent_vfolder_mount:
                if vfolder.name != ".logs":
                    continue
            mount = Mount(
                MountTypes.BIND,
                Path(vfolder.host_path),
                Path(vfolder.kernel_path),
                vfolder.mount_perm,
            )
            result.append(mount)
        return VFolderMountResult(mounts=result)

    @override
    async def teardown(self, resource: VFolderMountResult) -> None:
        pass


class VFolderMountStage(ProvisionStage[VFolderMountSpec, VFolderMountResult]):
    pass



import enum
import re
from collections.abc import Mapping
from dataclasses import dataclass
from decimal import Decimal
from pathlib import Path
from typing import Final, override

import pkg_resources
from cachetools import LRUCache, cached

from ai.backend.agent.kernel import match_distro_data
from ai.backend.agent.resources import ComputerContext, KernelResourceSpec, Mount
from ai.backend.agent.utils import get_arch_name
from ai.backend.common.stage.types import (
    ArgsSpecGenerator,
    Provisioner,
    ProvisionStage,
)
from ai.backend.common.types import (
    DeviceName,
    MountPermission,
    MountTypes,
)

DISTRO_PATTERN = re.compile(r"\.([a-z-]+\d+\.\d+)\.")
ARTIFACT_PATH = Path(pkg_resources.resource_filename("ai.backend.agent", "../runner"))


class LibcStyle(enum.StrEnum):
    GLIBC = "glibc"
    MUSL = "musl"


FALLBACK_KERNEL_PYTHON_VERSION: Final[str] = "3.6"  # Need to update?


@dataclass
class KernelRunnerInfo:
    distro: str
    architecture: str
    libc_style: LibcStyle
    krunner_volume: str
    krunner_py_version: str


@dataclass
class KernelRunnerMountSpec:
    distro: str
    krunner_volumes: Mapping[str, str]
    sandbox_type: str  # docker, jail

    existing_computers: Mapping[DeviceName, ComputerContext]
    resource_spec: KernelResourceSpec


class KernelRunnerMountSpecGenerator(ArgsSpecGenerator[KernelRunnerMountSpec]):
    pass


@dataclass
class KernelRunnerMountResult:
    mounts: list[Mount]


class KernelRunnerMountProvisioner(Provisioner[KernelRunnerMountSpec, KernelRunnerMountResult]):
    """
    Provisioner for the kernel creation setup stage.
    This is a no-op provisioner as it does not create any resources.
    """

    @property
    @override
    def name(self) -> str:
        return "docker-krunner-mount"

    @override
    async def setup(self, spec: KernelRunnerMountSpec) -> KernelRunnerMountResult:
        info = self._get_krunner_info(spec.distro, spec.krunner_volumes)
        mounts: list[Mount] = [
            *self._prepare_default_mounts(),
            *self._prepare_static_binary_mounts(info),
            *self._prepare_libbaihook_mounts(info),
            *self._prepare_krunner_volume_mounts(info),
            *self._prepare_python_lib_mounts(info),
            *self._prepare_musl_mounts(info),
            *self._prepare_jail_mounts(info, spec.sandbox_type),
        ]

        mounts.extend(await self._prepare_hook_mounts(spec, info))
        return KernelRunnerMountResult(
            mounts=mounts,
        )

    def _resolve_krunner_filepath(self, filename: str) -> Path:
        return Path(
            pkg_resources.resource_filename(
                "ai.backend.runner",
                "../" + filename,
            )
        ).resolve()

    def _find_artifacts(self, pattern: str) -> Mapping[str, str]:
        artifacts = {}
        for p in ARTIFACT_PATH.glob(pattern):
            m = DISTRO_PATTERN.search(p.name)
            if m is not None:
                artifacts[m.group(1)] = p.name
        return artifacts

    def _parse_mount(self, source: str, target: str) -> Mount:
        return Mount(
            MountTypes.BIND,
            self._resolve_krunner_filepath(source),
            Path(target),
            MountPermission.READ_ONLY,
        )

    def _prepare_default_mounts(self) -> list[Mount]:
        return [
            self._parse_mount("runner/extract_dotfiles.py", "/opt/kernel/extract_dotfiles.py"),
            self._parse_mount("runner/entrypoint.sh", "/opt/kernel/entrypoint.sh"),
            self._parse_mount("runner/fantompass.py", "/opt/kernel/fantompass.py"),
            self._parse_mount("runner/hash_phrase.py", "/opt/kernel/hash_phrase.py"),
            self._parse_mount("runner/words.json", "/opt/kernel/words.json"),
            self._parse_mount(
                "runner/DO_NOT_STORE_PERSISTENT_FILES_HERE.md",
                "/home/work/DO_NOT_STORE_PERSISTENT_FILES_HERE.md",
            ),
        ]

    def _prepare_musl_mounts(self, runner_info: KernelRunnerInfo) -> list[Mount]:
        """
        Mounts musl-specific files to the target path.
        """
        if runner_info.libc_style != LibcStyle.MUSL:
            return []
        return [
            Mount(
                MountTypes.BIND,
                self._resolve_krunner_filepath("runner/terminfo.alpine3.8"),
                Path("/home/work/.terminfo"),
                MountPermission.READ_ONLY,
            )
        ]

    def _prepare_static_binary_mounts(self, runner_info: KernelRunnerInfo) -> list[Mount]:
        """
        Mounts static binary files to the target path.
        """
        architecture = runner_info.architecture

        return [
            self._parse_mount(f"runner/su-exec.{architecture}.bin", "/opt/kernel/su-exec"),
            self._parse_mount(
                f"runner/dropbearmulti.{architecture}.bin", "/opt/kernel/dropbearmulti"
            ),
            self._parse_mount(f"runner/sftp-server.{architecture}.bin", "/opt/kernel/sftp-server"),
            self._parse_mount(f"runner/tmux.{architecture}.bin", "/opt/kernel/tmux"),
        ]

    def _prepare_libbaihook_mounts(self, runner_info: KernelRunnerInfo) -> list[Mount]:
        architecture = runner_info.architecture
        distro = runner_info.distro
        candidates = self._find_artifacts(f"libbaihook.*.{architecture}.so")
        _, cand = match_distro_data(candidates, distro)
        resolved_path = self._resolve_krunner_filepath(f"runner/{cand}")
        return [
            Mount(
                MountTypes.BIND,
                resolved_path,
                Path("/opt/kernel/libbaihook.so"),
                MountPermission.READ_ONLY,
            )
        ]

    def _prepare_jail_mounts(self, runner_info: KernelRunnerInfo, sandbox_type: str) -> list[Mount]:
        if sandbox_type != "jail":
            return []
        architecture = runner_info.architecture
        distro = runner_info.distro
        jail_candidates = self._find_artifacts(f"jail.*.{architecture}.bin")
        _, jail_cand = match_distro_data(jail_candidates, distro)
        resolved_path = self._resolve_krunner_filepath(f"runner/{jail_cand}")
        return [
            Mount(
                MountTypes.BIND,
                resolved_path,
                Path("/opt/kernel/jail"),
                MountPermission.READ_ONLY,
            )
        ]

    def _prepare_krunner_volume_mounts(
        self,
        runner_info: KernelRunnerInfo,
    ) -> list[Mount]:
        return [
            Mount(
                MountTypes.VOLUME,
                Path(runner_info.krunner_volume),
                Path("/opt/backend.ai"),
                MountPermission.READ_ONLY,
            )
        ]

    def _prepare_python_lib_mounts(
        self,
        runner_info: KernelRunnerInfo,
    ) -> list[Mount]:
        pylib_path = f"/opt/backend.ai/lib/python{runner_info.krunner_py_version}/site-packages/"
        kernel_pkg_path = self._resolve_krunner_filepath("kernel")
        helpers_pkg_path = self._resolve_krunner_filepath("helpers")
        return [
            Mount(
                MountTypes.BIND,
                kernel_pkg_path,
                Path(pylib_path + "ai/backend/kernel"),
                MountPermission.READ_ONLY,
            ),
            Mount(
                MountTypes.BIND,
                helpers_pkg_path,
                Path(pylib_path + "ai/backend/helpers"),
                MountPermission.READ_ONLY,
            ),
        ]

    async def _prepare_hook_mounts(
        self, spec: KernelRunnerMountSpec, runner_info: KernelRunnerInfo
    ) -> list[Mount]:
        mounts: list[Mount] = []
        already_injected_hooks: set[Path] = set()
        for dev_type, device_alloc in spec.resource_spec.allocations.items():
            computer_ctx = spec.existing_computers[dev_type]
            alloc_sum = Decimal(0)
            for per_dev_alloc in device_alloc.values():
                alloc_sum += sum(per_dev_alloc.values())
            do_hook_mount = alloc_sum > 0
            if do_hook_mount:
                hook_paths = await computer_ctx.instance.get_hooks(
                    runner_info.distro, runner_info.architecture
                )
                for hook_path in hook_paths:
                    if hook_path in already_injected_hooks:
                        continue
                    container_hook_path = f"/opt/kernel/{hook_path.name}"
                    already_injected_hooks.add(hook_path)
                    mounts.append(
                        Mount(
                            MountTypes.BIND,
                            hook_path,
                            Path(container_hook_path),
                            MountPermission.READ_ONLY,
                        )
                    )
        return mounts

    @cached(
        cache=LRUCache(maxsize=32),
        key=lambda self, distro, krunner_volumes: (distro, tuple(sorted(krunner_volumes.items()))),
    )
    def _get_krunner_info(
        self, distro: str, krunner_volumes: Mapping[str, str]
    ) -> KernelRunnerInfo:
        matched_distro, krunner_volume = match_distro_data(krunner_volumes, distro)
        matched_libc_style = LibcStyle.GLIBC
        if distro.startswith("alpine"):
            matched_libc_style = LibcStyle.MUSL
        krunner_pyver = FALLBACK_KERNEL_PYTHON_VERSION
        if m := re.search(r"^([a-z-]+)(\d+(\.\d+)*)?$", matched_distro):
            matched_distro_pkgname = m.group(1).replace("-", "_")
            try:
                krunner_pyver = (
                    Path(
                        pkg_resources.resource_filename(
                            f"ai.backend.krunner.{matched_distro_pkgname}",
                            f"krunner-python.{matched_distro}.txt",
                        )
                    )
                    .read_text()
                    .strip()
                )
            except FileNotFoundError:
                pass
        arch = get_arch_name()
        return KernelRunnerInfo(
            distro=matched_distro,
            architecture=arch,
            libc_style=matched_libc_style,
            krunner_volume=krunner_volume,
            krunner_py_version=krunner_pyver,
        )

    @override
    async def teardown(self, resource: KernelRunnerMountResult) -> None:
        pass


class KernelRunnerMountStage(ProvisionStage[KernelRunnerMountSpec, KernelRunnerMountResult]):
    pass



from __future__ import annotations

import asyncio
import functools
import importlib
import logging
import os
import shutil
import signal
import ssl
import sys
import time
from collections import OrderedDict, defaultdict
from ipaddress import IPv4Address, IPv6Address, ip_network
from pathlib import Path
from pprint import pformat, pprint
from typing import (
    TYPE_CHECKING,
    Any,
    AsyncGenerator,
    Callable,
    ClassVar,
    Coroutine,
    Iterable,
    Literal,
    Mapping,
    Optional,
    Sequence,
    Set,
    cast,
)
from uuid import UUID

import aiohttp_cors
import aiomonitor
import aiotools
import click
import tomlkit
from aiohttp import web
from aiotools import aclosing
from callosum.lower.zeromq import ZeroMQAddress, ZeroMQRPCTransport
from callosum.ordering import ExitOrderedAsyncScheduler
from callosum.rpc import Peer, RPCMessage
from etcd_client import WatchEventType
from setproctitle import setproctitle
from zmq.auth.certs import load_certificate

from ai.backend.agent.metrics.metric import RPCMetricObserver
from ai.backend.agent.resources import scan_gpu_alloc_map
from ai.backend.agent.stats import StatModes
from ai.backend.common import config, identity, msgpack, utils
from ai.backend.common.auth import AgentAuthHandler, PublicKey, SecretKey
from ai.backend.common.bgtask.bgtask import ProgressReporter
from ai.backend.common.configs.redis import RedisConfig
from ai.backend.common.defs import RedisRole
from ai.backend.common.docker import ImageRef
from ai.backend.common.dto.agent.response import (
    AbstractAgentResp,
    CodeCompletionResp,
    DropKernelRegistryResp,
    PurgeContainersResp,
    PurgeImagesResp,
)
from ai.backend.common.dto.manager.rpc_request import PurgeImagesReq
from ai.backend.common.etcd import AsyncEtcd, ConfigScopes
from ai.backend.common.events.event_types.kernel.anycast import (
    KernelTerminatedAnycastEvent,
)
from ai.backend.common.events.event_types.kernel.broadcast import (
    KernelTerminatedBroadcastEvent,
)
from ai.backend.common.events.event_types.kernel.types import KernelLifecycleEventReason
from ai.backend.common.exception import ConfigurationError
from ai.backend.common.json import pretty_json
from ai.backend.common.metrics.http import (
    build_api_metric_middleware,
    build_prometheus_metrics_handler,
)
from ai.backend.common.metrics.metric import CommonMetricRegistry
from ai.backend.common.metrics.profiler import Profiler, PyroscopeArgs
from ai.backend.common.service_discovery.etcd_discovery.service_discovery import (
    ETCDServiceDiscovery,
    ETCDServiceDiscoveryArgs,
)
from ai.backend.common.service_discovery.redis_discovery.service_discovery import (
    RedisServiceDiscovery,
    RedisServiceDiscoveryArgs,
)
from ai.backend.common.service_discovery.service_discovery import (
    ServiceDiscovery,
    ServiceDiscoveryLoop,
    ServiceEndpoint,
    ServiceMetadata,
)
from ai.backend.common.types import (
    ClusterInfo,
    CommitStatus,
    ContainerId,
    ContainerKernelId,
    HardwareMetadata,
    HostPortPair,
    ImageConfig,
    ImageRegistry,
    KernelCreationConfig,
    KernelId,
    QueueSentinel,
    ServiceDiscoveryType,
    SessionId,
    aobject,
)
from ai.backend.common.utils import current_loop
from ai.backend.logging import BraceStyleAdapter, Logger, LogLevel
from ai.backend.logging.otel import OpenTelemetrySpec

from . import __version__ as VERSION
from .config.unified import (
    AgentUnifiedConfig,
    APIConfig,
    ContainerLogsConfig,
    ContainerSandboxType,
    DockerExtraConfig,
    EventLoopType,
    KernelLifecyclesConfig,
    ScratchType,
)
from .exception import ResourceError
from .monitor import AgentErrorPluginContext, AgentStatsPluginContext
from .types import (
    AgentBackend,
    KernelLifecycleStatus,
    KernelOwnershipData,
    LifecycleEvent,
    VolumeInfo,
)
from .utils import get_arch_name, get_subnet_ip

if TYPE_CHECKING:
    from .agent import AbstractAgent

log = BraceStyleAdapter(logging.getLogger(__spec__.name))

deeplearning_image_keys = {
    "tensorflow",
    "caffe",
    "keras",
    "torch",
    "mxnet",
    "theano",
}

deeplearning_sample_volume = VolumeInfo(
    "deeplearning-samples",
    "/home/work/samples",
    "ro",
)

agent_instance: AgentRPCServer


async def get_extra_volumes(docker, lang):
    avail_volumes = (await docker.volumes.list())["Volumes"]
    if not avail_volumes:
        return []
    avail_volume_names = set(v["Name"] for v in avail_volumes)

    # deeplearning specialization
    # TODO: extract as config
    volume_list = []
    for k in deeplearning_image_keys:
        if k in lang:
            volume_list.append(deeplearning_sample_volume)
            break

    # Mount only actually existing volumes
    mount_list = []
    for vol in volume_list:
        if vol.name in avail_volume_names:
            mount_list.append(vol)
        else:
            log.info(
                "skipped attaching extra volume {0} to a kernel based on image {1}",
                vol.name,
                lang,
            )
    return mount_list


def collect_error(meth: Callable) -> Callable:
    @functools.wraps(meth)
    async def _inner(self: AgentRPCServer, *args, **kwargs):
        try:
            return await meth(self, *args, **kwargs)
        except Exception:
            await self.agent.produce_error_event()
            raise

    return _inner


class RPCFunctionRegistry:
    functions: Set[str]
    _metric_observer: RPCMetricObserver

    def __init__(self) -> None:
        self.functions = set()
        self._metric_observer = RPCMetricObserver.instance()

    def __call__(
        self,
        meth: Callable[..., Coroutine[None, None, Any]],
    ) -> Callable[[AgentRPCServer, RPCMessage], Coroutine[None, None, Any]]:
        @functools.wraps(meth)
        @_collect_metrics(self._metric_observer, meth.__name__)
        async def _inner(self_: AgentRPCServer, request: RPCMessage) -> Any:
            try:
                if request.body is None:
                    return await meth(self_)
                else:
                    return await meth(
                        self_,
                        *request.body["args"],
                        **request.body["kwargs"],
                    )
            except (asyncio.CancelledError, asyncio.TimeoutError):
                raise
            except ResourceError:
                # This is an expected scenario.
                raise
            except Exception:
                log.exception("unexpected error")
                await self_.error_monitor.capture_exception()
                raise

        self.functions.add(meth.__name__)
        return _inner


class RPCFunctionRegistryV2:
    functions: Set[str]
    _metric_observer: RPCMetricObserver

    def __init__(self) -> None:
        self.functions = set()
        self._metric_observer = RPCMetricObserver.instance()

    def __call__(
        self,
        meth: Callable[..., Coroutine[None, None, AbstractAgentResp]],
    ) -> Callable[[AgentRPCServer, RPCMessage], Coroutine[None, None, Any]]:
        @functools.wraps(meth)
        @_collect_metrics(self._metric_observer, meth.__name__)
        async def _inner(self_: AgentRPCServer, request: RPCMessage) -> Any:
            try:
                if request.body is None:
                    return await meth(self_)
                else:
                    res = await meth(
                        self_,
                        *request.body["args"],
                        **request.body["kwargs"],
                    )
                    return res.as_dict()
            except (asyncio.CancelledError, asyncio.TimeoutError):
                raise
            except ResourceError:
                # This is an expected scenario.
                raise
            except Exception:
                log.exception("unexpected error")
                await self_.error_monitor.capture_exception()
                raise

        self.functions.add(meth.__name__)
        return _inner


def _collect_metrics(observer: RPCMetricObserver, method_name: str) -> Callable:
    def decorator(meth: Callable) -> Callable[[AgentRPCServer, RPCMessage], Any]:
        @functools.wraps(meth)
        async def _inner(self: AgentRPCServer, *args, **kwargs) -> Any:
            start_time = time.perf_counter()
            try:
                res = await meth(self, *args, **kwargs)
                duration = time.perf_counter() - start_time
                observer.observe_rpc_request_success(
                    method=method_name,
                    duration=duration,
                )
                return res
            except BaseException as e:
                duration = time.perf_counter() - start_time
                observer.observe_rpc_request_failure(
                    method=method_name,
                    duration=duration,
                    exception=e,
                )
                raise

        return _inner

    return decorator


class AgentRPCServer(aobject):
    rpc_function: ClassVar[RPCFunctionRegistry] = RPCFunctionRegistry()
    rpc_function_v2: ClassVar[RPCFunctionRegistryV2] = RPCFunctionRegistryV2()

    rpc_auth_manager_public_key: Optional[PublicKey]
    rpc_auth_agent_public_key: Optional[PublicKey]
    rpc_auth_agent_secret_key: Optional[SecretKey]

    loop: asyncio.AbstractEventLoop
    agent: AbstractAgent
    rpc_server: Peer
    rpc_addr: str
    agent_addr: str

    _stop_signal: signal.Signals
    debug_server_task: asyncio.Task

    def __init__(
        self,
        etcd: AsyncEtcd,
        local_config: AgentUnifiedConfig,
        *,
        skip_detect_manager: bool = False,
    ) -> None:
        self.loop = current_loop()
        self.etcd = etcd
        self.local_config = local_config
        self.skip_detect_manager = skip_detect_manager
        self._stop_signal = signal.SIGTERM

    async def __ainit__(self) -> None:
        # Start serving requests.
        await self.update_status("starting")

        if not self.skip_detect_manager:
            await self.detect_manager()

        await self.read_agent_config()
        await self.read_agent_config_container()

        self.stats_monitor = AgentStatsPluginContext(
            self.etcd, self.local_config.model_dump(by_alias=True)
        )
        self.error_monitor = AgentErrorPluginContext(
            self.etcd, self.local_config.model_dump(by_alias=True)
        )
        await self.stats_monitor.init()
        await self.error_monitor.init()

        if self.local_config.agent.rpc_auth_agent_keypair is not None:
            manager_pkey, _ = load_certificate(
                str(self.local_config.agent.rpc_auth_manager_public_key)
            )
            self.rpc_auth_manager_public_key = PublicKey(manager_pkey)
            agent_pkey, agent_skey = load_certificate(
                str(self.local_config.agent.rpc_auth_agent_keypair)
            )
            assert agent_skey is not None
            self.rpc_auth_agent_public_key = PublicKey(agent_pkey)
            self.rpc_auth_agent_secret_key = SecretKey(agent_skey)
            log.info(
                "RPC encryption and authentication is enabled. "
                "(agent_public_key = '{}', manager_public_key='{}')",
                self.rpc_auth_agent_public_key.decode("ascii"),
                self.rpc_auth_manager_public_key.decode("ascii"),
            )
            auth_handler = AgentAuthHandler(
                "local",
                self.rpc_auth_manager_public_key,
                self.rpc_auth_agent_public_key,
                self.rpc_auth_agent_secret_key,
            )
        else:
            self.rpc_auth_manager_public_key = None
            self.rpc_auth_agent_public_key = None
            self.rpc_auth_agent_secret_key = None
            auth_handler = None

        backend = self.local_config.agent.backend
        agent_mod = importlib.import_module(f"ai.backend.agent.{backend.value}")
        self.agent = await agent_mod.get_agent_cls().new(  # type: ignore
            self.etcd,
            self.local_config,
            stats_monitor=self.stats_monitor,
            error_monitor=self.error_monitor,
            agent_public_key=self.rpc_auth_agent_public_key,
        )

        rpc_addr = self.local_config.agent.rpc_listen_addr
        self.rpc_server = Peer(
            bind=ZeroMQAddress(f"tcp://{rpc_addr.address}"),
            transport=ZeroMQRPCTransport,
            authenticator=auth_handler,
            scheduler=ExitOrderedAsyncScheduler(),
            serializer=msgpack.packb,
            deserializer=msgpack.unpackb,
            debug_rpc=self.local_config.debug.enabled,
        )
        for func_name in self.rpc_function.functions:
            self.rpc_server.handle_function(func_name, getattr(self, func_name))

        for func_name in self.rpc_function_v2.functions:
            self.rpc_server.handle_function(func_name, getattr(self, func_name))

        log.info("started handling RPC requests at {}", rpc_addr)

        debug_socket_path = self.local_config.agent.ipc_base_path / "agent-registry-snapshot.sock"
        server = await asyncio.start_unix_server(
            self.status_snapshot_request_handler, debug_socket_path.as_posix()
        )

        async def _debug_server_task():
            try:
                async with server:
                    await server.serve_forever()
            except Exception:
                log.exception("_debug_server_task():")
                raise

        self.debug_server_task = asyncio.create_task(_debug_server_task())

        await self.etcd.put("ip", rpc_addr.host, scope=ConfigScopes.NODE)

        watcher_port = utils.nmget(
            self.local_config.model_dump(), "watcher.service-addr.port", None
        )
        if watcher_port is not None:
            await self.etcd.put("watcher_port", watcher_port, scope=ConfigScopes.NODE)

        await self.update_status("running")

    async def status_snapshot_request_handler(
        self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter
    ) -> None:
        def _ensure_serializable(o) -> Any:
            match o:
                case dict() | defaultdict() | OrderedDict():
                    return {_ensure_serializable(k): _ensure_serializable(v) for k, v in o.items()}
                case list():
                    return [_ensure_serializable(e) for e in o]
                case set():
                    return set([_ensure_serializable(e) for e in o])
                case tuple():
                    return tuple([_ensure_serializable(e) for e in o])
                case _:
                    return str(o)

        try:
            if self.agent:
                snapshot = {
                    "registry": {
                        str(kern_id): _ensure_serializable(kern.__getstate__())
                        for kern_id, kern in self.agent.kernel_registry.items()
                    },
                    "allocs": {
                        str(computer): _ensure_serializable(
                            dict(computer_ctx.alloc_map.allocations)
                        )
                        for computer, computer_ctx in self.agent.computers.items()
                    },
                }
                writer.write(pretty_json(snapshot))
            await writer.drain()
            writer.close()
            await writer.wait_closed()
        except Exception:
            log.exception("status_snapshot_request_handler():")
            raise

    async def detect_manager(self):
        log.info("detecting the manager...")
        manager_instances = await self.etcd.get_prefix("nodes/manager")
        if not manager_instances:
            log.warning("watching etcd to wait for the manager being available")
            async with aclosing(self.etcd.watch_prefix("nodes/manager")) as agen:
                async for ev in agen:
                    match ev:
                        case QueueSentinel.CLOSED | QueueSentinel.TIMEOUT:
                            break
                        case _:
                            if ev.event == WatchEventType.PUT and ev.value == "up":
                                break
        log.info("detected at least one manager running")

    async def read_agent_config(self):
        # Fill up Redis configs from etcd and store as separate attributes
        self._redis_config = config.redis_config_iv.check(
            await self.etcd.get_prefix("config/redis"),
        )
        log.info("configured redis: {0}", self._redis_config)

        # Update local_config with redis settings
        # Convert HostPortPair to dict format for compatibility
        redis_config_dict = self._redis_config.copy()
        if isinstance(self._redis_config.get("addr"), object):
            addr = self._redis_config["addr"]
            if addr is not None:
                redis_config_dict["addr"] = f"{addr.host}:{addr.port}"

        redis_config = RedisConfig.model_validate(redis_config_dict)
        self.local_config.redis = redis_config

        # Fill up vfolder configs from etcd and store as separate attributes
        # TODO: Integrate vfolder_config into local_config
        self._vfolder_config = config.vfolder_config_iv.check(
            await self.etcd.get_prefix("volumes"),
        )
        if self._vfolder_config["mount"] is None:
            log.info(
                "assuming use of storage-proxy since vfolder mount path is not configured in etcd"
            )
        else:
            log.info("configured vfolder mount base: {0}", self._vfolder_config["mount"])
            log.info("configured vfolder fs prefix: {0}", self._vfolder_config["fsprefix"])

        # Fill up shared agent configurations from etcd.
        agent_etcd_config_raw = await self.etcd.get_prefix("config/agent")
        if agent_etcd_config_raw:
            try:
                # Parse specific etcd configs and update the unified config
                api_config = APIConfig.model_validate(agent_etcd_config_raw.get("api", {}))
                container_logs_config = ContainerLogsConfig.model_validate(
                    agent_etcd_config_raw.get("container-logs", {})
                )
                kernel_lifecycles_config = KernelLifecyclesConfig.model_validate(
                    agent_etcd_config_raw.get("kernel-lifecycles", {})
                )

                # Update local config with parsed values
                self.local_config = self.local_config.model_copy(
                    update={
                        "api": api_config,
                        "container_logs": container_logs_config,
                        "kernel_lifecycles": kernel_lifecycles_config,
                    }
                )
            except Exception as e:
                log.warning("etcd: agent-config error: {}", e)

    async def read_agent_config_container(self):
        # Fill up global container configurations from etcd.
        try:
            container_etcd_config_raw = await self.etcd.get_prefix("config/container")
            if container_etcd_config_raw:
                # Update config by creating a new instance with modified values
                container_updates = {}
                if "kernel-uid" in container_etcd_config_raw:
                    container_updates["kernel_uid"] = container_etcd_config_raw["kernel-uid"]
                    log.info(
                        "etcd: container-config: kernel-uid={}".format(
                            container_etcd_config_raw["kernel-uid"]
                        )
                    )
                if "kernel-gid" in container_etcd_config_raw:
                    container_updates["kernel_gid"] = container_etcd_config_raw["kernel-gid"]
                    log.info(
                        "etcd: container-config: kernel-gid={}".format(
                            container_etcd_config_raw["kernel-gid"]
                        )
                    )

                if container_updates:
                    # Create new container config with updated values
                    new_container_config = self.local_config.container.model_copy(
                        update=container_updates
                    )
                    # Create new full config with updated container section
                    self.local_config = self.local_config.model_copy(
                        update={"container": new_container_config}
                    )
        except Exception as e:
            log.warning("etcd: container-config error: {}".format(e))

    async def __aenter__(self) -> None:
        await self.rpc_server.__aenter__()

    def mark_stop_signal(self, stop_signal: signal.Signals) -> None:
        self._stop_signal = stop_signal

    async def __aexit__(self, *exc_info) -> None:
        # Stop receiving further requests.
        await self.rpc_server.__aexit__(*exc_info)
        self.debug_server_task.cancel()
        await asyncio.sleep(0)
        if not self.debug_server_task.done():
            await self.debug_server_task
        await self.agent.shutdown(self._stop_signal)
        await self.stats_monitor.cleanup()
        await self.error_monitor.cleanup()

    @collect_error
    async def update_status(self, status):
        await self.etcd.put("", status, scope=ConfigScopes.NODE)

    @rpc_function
    @collect_error
    async def update_scaling_group(self, scaling_group):
        cfg_src_path = config.find_config_file("agent")
        with open(cfg_src_path, "r") as f:
            data = tomlkit.load(f)
            data["agent"]["scaling-group"] = scaling_group
        shutil.copy(cfg_src_path, f"{cfg_src_path}.bak")
        with open(cfg_src_path, "w") as f:
            tomlkit.dump(data, f)
        # Update local config with new scaling group
        new_agent_config = self.local_config.agent.model_copy(
            update={"scaling_group": scaling_group}
        )
        self.local_config = self.local_config.model_copy(update={"agent": new_agent_config})
        log.info("rpc::update_scaling_group()")

    @rpc_function
    @collect_error
    async def ping(self, msg: str) -> str:
        log.debug("rpc::ping()")
        return msg

    @rpc_function
    @collect_error
    async def gather_hwinfo(self) -> Mapping[str, HardwareMetadata]:
        log.debug("rpc::gather_hwinfo()")
        return await self.agent.gather_hwinfo()

    @rpc_function
    @collect_error
    async def ping_kernel(self, kernel_id: str) -> dict[str, float] | None:
        log.debug("rpc::ping_kernel(k:{})", kernel_id)
        return await self.agent.ping_kernel(KernelId(UUID(kernel_id)))

    @rpc_function
    @collect_error
    async def check_pulling(self, image_name: str) -> bool:
        """Check if an image is being pulled."""
        log.debug("rpc::check_pulling(image:{})", image_name)
        return image_name in self.agent._active_pulls

    @rpc_function
    @collect_error
    async def check_creating(self, kernel_id: str) -> bool:
        """Check if a kernel is being created or already exists."""
        log.debug("rpc::check_creating(k:{})", kernel_id)
        kid = KernelId(UUID(kernel_id))
        # Check if kernel is being created OR already exists in registry
        return kid in self.agent._active_creates or kid in self.agent.kernel_registry

    @rpc_function
    @collect_error
    async def check_running(self, kernel_id: str) -> bool:
        """Check if a kernel is running."""
        log.debug("rpc::check_running(k:{})", kernel_id)
        kid = KernelId(UUID(kernel_id))

        # Safely get kernel from registry
        kernel_obj = self.agent.kernel_registry.get(kid)

        # Check if kernel exists and is running
        if kernel_obj is None:
            return False

        return kernel_obj.state == KernelLifecycleStatus.RUNNING

    @rpc_function
    @collect_error
    async def sync_kernel_registry(
        self,
        raw_kernel_session_ids: Iterable[tuple[str, str]],
    ) -> None:
        kernel_session_ids = [
            (KernelId(UUID(raw_kid)), SessionId(UUID(raw_sid)))
            for raw_kid, raw_sid in raw_kernel_session_ids
        ]
        for kid, sid in kernel_session_ids:
            if kid not in self.agent.kernel_registry:
                # produce KernelTerminatedEvent
                await self.agent.anycast_and_broadcast_event(
                    KernelTerminatedAnycastEvent(
                        kid,
                        sid,
                        reason=KernelLifecycleEventReason.ALREADY_TERMINATED,
                    ),
                    KernelTerminatedBroadcastEvent(
                        kid,
                        sid,
                        reason=KernelLifecycleEventReason.ALREADY_TERMINATED,
                    ),
                )

        kernel_ids = {kern_id for kern_id, sess_id in kernel_session_ids}
        for kid, kernel in self.agent.kernel_registry.items():
            if kid not in kernel_ids:
                # destroy kernel
                await self.agent.inject_container_lifecycle_event(
                    kid,
                    kernel.session_id,
                    LifecycleEvent.DESTROY,
                    KernelLifecycleEventReason.NOT_FOUND_IN_MANAGER,
                    suppress_events=True,
                )

    @rpc_function
    @collect_error
    async def check_and_pull(
        self,
        image_configs: Mapping[str, ImageConfig],
    ) -> dict[str, str]:
        """
        Check whether the agent has images and pull if needed.
        Delegates to agent's check_and_pull method which handles tracking.
        """
        log.debug("rpc::check_and_pull(images:{})", list(image_configs.keys()))
        return await self.agent.check_and_pull(image_configs)

    @rpc_function
    @collect_error
    async def create_kernels(
        self,
        raw_session_id: str,
        raw_kernel_ids: Sequence[str],
        raw_configs: Sequence[dict],
        raw_cluster_info: dict,
        kernel_image_refs: dict[KernelId, ImageRef],
    ):
        cluster_info = cast(ClusterInfo, raw_cluster_info)
        session_id = SessionId(UUID(raw_session_id))
        coros = []
        throttle_sema = asyncio.Semaphore(self.local_config.agent.kernel_creation_concurrency)
        for raw_kernel_id, raw_config in zip(raw_kernel_ids, raw_configs):
            log.info(
                "rpc::create_kernel(k:{0}, img:{1})",
                raw_kernel_id,
                raw_config["image"]["canonical"],
            )
            kernel_id = KernelId(UUID(raw_kernel_id))
            kernel_config = cast(KernelCreationConfig, raw_config)
            coros.append(
                self.agent.create_kernel(
                    KernelOwnershipData(
                        kernel_id,
                        session_id,
                        self.agent.id,
                        raw_config.get("owner_user_id"),
                        raw_config.get("owner_project_id"),
                    ),
                    kernel_image_refs[kernel_id],
                    kernel_config,
                    cluster_info,
                    throttle_sema=throttle_sema,
                )
            )
        results = await asyncio.gather(*coros, return_exceptions=True)
        raw_results = []
        errors = []
        for result in results:
            match result:
                case BaseException():
                    errors.append(result)
                case _:
                    raw_results.append({
                        "id": str(result["id"]),
                        "kernel_host": result["kernel_host"],
                        "repl_in_port": result["repl_in_port"],
                        "repl_out_port": result["repl_out_port"],
                        "stdin_port": result["stdin_port"],  # legacy
                        "stdout_port": result["stdout_port"],  # legacy
                        "service_ports": result["service_ports"],
                        "container_id": result["container_id"],
                        "resource_spec": result["resource_spec"],
                        "attached_devices": result["attached_devices"],
                        "agent_addr": result["agent_addr"],
                        "scaling_group": result["scaling_group"],
                    })
        if errors:
            # Raise up the first error.
            if len(errors) == 1:
                raise errors[0]
            raise aiotools.TaskGroupError("agent.create_kernels() failed", errors)
        return raw_results

    @rpc_function
    @collect_error
    async def destroy_kernel(
        self,
        kernel_id: str,
        session_id: str,
        reason: Optional[KernelLifecycleEventReason] = None,
        suppress_events: bool = False,
    ):
        loop = asyncio.get_running_loop()
        done = loop.create_future()
        log.info("rpc::destroy_kernel(k:{0})", kernel_id)
        await self.agent.inject_container_lifecycle_event(
            KernelId(UUID(kernel_id)),
            SessionId(UUID(session_id)),
            LifecycleEvent.DESTROY,
            reason or KernelLifecycleEventReason.USER_REQUESTED,
            done_future=done,
            suppress_events=suppress_events,
        )
        return await done

    @rpc_function_v2
    @collect_error
    async def purge_containers(
        self,
        container_kernel_ids: list[tuple[str, str]],
    ) -> PurgeContainersResp:
        str_kernel_ids = [str(kid) for _, kid in container_kernel_ids]
        log.info("rpc::purge_containers(kernel_ids:{0})", str_kernel_ids)
        kernel_container_pairs = [
            ContainerKernelId(
                ContainerId(cid),
                KernelId(UUID(kid)),
            )
            for cid, kid in container_kernel_ids
        ]
        asyncio.create_task(self.agent.purge_containers(kernel_container_pairs))
        return PurgeContainersResp()

    @rpc_function_v2
    @collect_error
    async def drop_kernel_registry(
        self,
        kernel_ids: list[UUID],
    ) -> DropKernelRegistryResp:
        str_kernel_ids = [str(kid) for kid in kernel_ids]
        log.info("rpc::drop_kernel_registry(kernel_ids:{0})", str_kernel_ids)
        kernel_ids_to_purge = [KernelId(kid) for kid in kernel_ids]
        asyncio.create_task(self.agent.clean_kernel_objects(kernel_ids_to_purge))
        return DropKernelRegistryResp()

    @rpc_function
    @collect_error
    async def interrupt_kernel(self, kernel_id: str):
        log.info("rpc::interrupt_kernel(k:{0})", kernel_id)
        await self.agent.interrupt_kernel(KernelId(UUID(kernel_id)))

    @rpc_function_v2
    @collect_error
    async def get_completions(self, kernel_id: str, text: str, opts: dict) -> CodeCompletionResp:
        log.debug("rpc::get_completions(k:{0}, ...)", kernel_id)
        return await self.agent.get_completions(KernelId(UUID(kernel_id)), text, opts)

    @rpc_function
    @collect_error
    async def get_logs(self, kernel_id: str):
        log.info("rpc::get_logs(k:{0})", kernel_id)
        return await self.agent.get_logs(KernelId(UUID(kernel_id)))

    @rpc_function
    @collect_error
    async def restart_kernel(
        self,
        session_id: str,
        kernel_id: str,
        kernel_image: ImageRef,
        updated_config: dict,
    ) -> dict[str, Any]:
        log.info("rpc::restart_kernel(s:{0}, k:{1})", session_id, kernel_id)
        return await self.agent.restart_kernel(
            KernelOwnershipData(
                KernelId(UUID(kernel_id)),
                SessionId(UUID(session_id)),
                self.agent.id,
            ),
            kernel_image,
            cast(KernelCreationConfig, updated_config),
        )

    @rpc_function
    @collect_error
    async def execute(
        self,
        session_id: str,
        kernel_id: str,
        api_version: int,
        run_id: str,
        mode: Literal["query", "batch", "continue", "input"],
        code: str,
        opts: dict[str, Any],
        flush_timeout: float,
    ) -> dict[str, Any]:
        if mode != "continue":
            log.info(
                "rpc::execute(k:{0}, run-id:{1}, mode:{2}, code:{3!r})",
                kernel_id,
                run_id,
                mode,
                code[:20] + "..." if len(code) > 20 else code,
            )
        result = await self.agent.execute(
            SessionId(UUID(session_id)),
            KernelId(UUID(kernel_id)),
            run_id,
            mode,
            code,
            opts=opts,
            api_version=api_version,
            flush_timeout=flush_timeout,
        )
        return result

    @rpc_function
    @collect_error
    async def trigger_batch_execution(
        self,
        session_id: str,
        kernel_id: str,
        code: str,
        timeout: Optional[float],
    ) -> None:
        log.info(
            "rpc::trigger_batch_execution(k:{0}, s:{1}, code:{2}, timeout:{3})",
            kernel_id,
            session_id,
            code,
            timeout,
        )
        await self.agent.create_batch_execution_task(
            SessionId(UUID(session_id)), KernelId(UUID(kernel_id)), code, timeout
        )

    @rpc_function
    @collect_error
    async def start_service(
        self,
        kernel_id: str,
        service: str,
        opts: dict[str, Any],
    ) -> dict[str, Any]:
        log.info("rpc::start_service(k:{0}, app:{1})", kernel_id, service)
        return await self.agent.start_service(KernelId(UUID(kernel_id)), service, opts)

    @rpc_function
    @collect_error
    async def get_commit_status(
        self,
        kernel_id: str,
        subdir: str,
    ) -> dict[str, Any]:
        # Only this function logs debug since web sends request at short intervals
        log.debug("rpc::get_commit_status(k:{})", kernel_id)
        status: CommitStatus = await self.agent.get_commit_status(
            KernelId(UUID(kernel_id)),
            subdir,
        )
        return {
            "kernel": kernel_id,
            "status": status.value,
        }

    @rpc_function
    @collect_error
    async def commit(
        self,
        kernel_id: str,
        subdir: str,
        *,
        canonical: str | None = None,
        filename: str | None = None,
        extra_labels: dict[str, str] = {},
    ) -> dict[str, Any]:
        log.info("rpc::commit(k:{})", kernel_id)
        bgtask_mgr = self.agent.background_task_manager

        async def _commit(reporter: ProgressReporter) -> None:
            await self.agent.commit(
                reporter,
                KernelId(UUID(kernel_id)),
                subdir,
                canonical=canonical,
                filename=filename,
                extra_labels=extra_labels,
            )

        task_id = await bgtask_mgr.start(_commit)
        return {
            "bgtask_id": str(task_id),
            "kernel": kernel_id,
            "path": str(Path(subdir, filename)) if filename else None,
        }

    @rpc_function
    @collect_error
    async def push_image(
        self,
        image_ref: ImageRef,
        registry_conf: ImageRegistry,
    ) -> dict[str, Any]:
        log.info("rpc::push_image(c:{})", image_ref.canonical)
        bgtask_mgr = self.agent.background_task_manager

        image_push_timeout = cast(Optional[float], self.local_config.api.push_timeout)

        async def _push_image(reporter: ProgressReporter) -> None:
            await self.agent.push_image(
                image_ref,
                registry_conf,
                timeout=image_push_timeout,
            )

        task_id = await bgtask_mgr.start(_push_image)
        return {
            "bgtask_id": str(task_id),
            "canonical": image_ref.canonical,
        }

    @rpc_function_v2
    @collect_error
    async def purge_images(
        self, image_canonicals: list[str], force: bool, noprune: bool
    ) -> PurgeImagesResp:
        log.info(
            "rpc::purge_images(images:{0}, force:{1}, noprune:{2})",
            image_canonicals,
            force,
            noprune,
        )
        return await self.agent.purge_images(
            PurgeImagesReq(images=image_canonicals, force=force, noprune=noprune)
        )

    @rpc_function
    @collect_error
    async def get_local_config(self) -> Mapping[str, Any]:
        report_path: Path | None = self.local_config.agent.abuse_report_path
        return {
            "agent": {
                "abuse-report-path": str(report_path) if report_path is not None else "",
            },
            "watcher": getattr(self, "_watcher_config", {}),
        }

    @rpc_function
    @collect_error
    async def shutdown_service(
        self,
        kernel_id,  # type: str
        service,  # type: str
    ):
        log.info("rpc::shutdown_service(k:{0}, app:{1})", kernel_id, service)
        return await self.agent.shutdown_service(KernelId(UUID(kernel_id)), service)

    @rpc_function
    @collect_error
    async def upload_file(self, kernel_id: str, filename: str, filedata: bytes):
        log.info("rpc::upload_file(k:{0}, fn:{1})", kernel_id, filename)
        await self.agent.accept_file(KernelId(UUID(kernel_id)), filename, filedata)

    @rpc_function
    @collect_error
    async def download_file(self, kernel_id: str, filepath: str):
        log.info("rpc::download_file(k:{0}, fn:{1})", kernel_id, filepath)
        return await self.agent.download_file(KernelId(UUID(kernel_id)), filepath)

    @rpc_function
    @collect_error
    async def download_single(self, kernel_id: str, filepath: str):
        log.info("rpc::download_single(k:{0}, fn:{1})", kernel_id, filepath)
        return await self.agent.download_single(KernelId(UUID(kernel_id)), filepath)

    @rpc_function
    @collect_error
    async def list_files(self, kernel_id: str, path: str):
        log.info("rpc::list_files(k:{0}, fn:{1})", kernel_id, path)
        return await self.agent.list_files(KernelId(UUID(kernel_id)), path)

    @rpc_function
    @collect_error
    async def shutdown_agent(self, terminate_kernels: bool):
        # TODO: implement
        log.info("rpc::shutdown_agent()")
        pass

    @rpc_function
    @collect_error
    async def create_local_network(self, network_name: str) -> None:
        log.debug("rpc::create_local_network(name:{})", network_name)
        return await self.agent.create_local_network(network_name)

    @rpc_function
    @collect_error
    async def destroy_local_network(self, network_name: str) -> None:
        log.debug("rpc::destroy_local_network(name:{})", network_name)
        return await self.agent.destroy_local_network(network_name)

    @rpc_function
    @collect_error
    async def reset_agent(self):
        log.debug("rpc::reset()")
        kernel_ids = tuple(self.agent.kernel_registry.keys())
        tasks = []
        for kernel_id in kernel_ids:
            try:
                task = asyncio.ensure_future(self.agent.destroy_kernel(kernel_id, "agent-reset"))
                tasks.append(task)
            except Exception:
                await self.error_monitor.capture_exception()
                log.exception("reset: destroying {0}", kernel_id)
        await asyncio.gather(*tasks)

    @rpc_function
    @collect_error
    async def assign_port(self):
        log.debug("rpc::assign_port()")
        return self.agent.port_pool.pop()

    @rpc_function
    @collect_error
    async def release_port(self, port_no: int):
        log.debug("rpc::release_port(port_no:{})", port_no)
        self.agent.port_pool.add(port_no)

    @rpc_function
    @collect_error
    async def scan_gpu_alloc_map(self) -> Mapping[str, Any]:
        log.debug("rpc::scan_gpu_alloc_map()")
        scratch_root = self.agent.local_config.container.scratch_root
        result = await scan_gpu_alloc_map(list(self.agent.kernel_registry.keys()), scratch_root)
        return {k: str(v) for k, v in result.items()}


@aiotools.server_context
async def server_main_logwrapper(
    loop: asyncio.AbstractEventLoop,
    pidx: int,
    _args: Sequence[Any],
) -> AsyncGenerator[None, signal.Signals]:
    setproctitle(f"backend.ai: agent worker-{pidx}")
    local_cfg: AgentUnifiedConfig = _args[0]
    log_endpoint = _args[1]
    logger = Logger(
        local_cfg.logging,
        is_master=False,
        log_endpoint=log_endpoint,
        msgpack_options={
            "pack_opts": msgpack.DEFAULT_PACK_OPTS,
            "unpack_opts": msgpack.DEFAULT_UNPACK_OPTS,
        },
    )
    with logger:
        async with server_main(loop, pidx, _args):
            yield


async def check_health(request: web.Request) -> web.Response:
    """Simple health check endpoint"""
    from . import __version__
    from .types import HealthResponse

    request["do_not_print_access_log"] = True

    response = HealthResponse(
        status="healthy",
        version=__version__,
        component="agent",
    )
    return web.json_response(response.model_dump())


def build_root_server() -> web.Application:
    metric_registry = CommonMetricRegistry.instance()
    app = web.Application(
        middlewares=[
            build_api_metric_middleware(metric_registry.api),
        ],
    )
    cors = aiohttp_cors.setup(
        app,
        defaults={
            "*": aiohttp_cors.ResourceOptions(
                allow_credentials=False, expose_headers="*", allow_headers="*"
            ),
        },
    )
    cors.add(app.router.add_route("GET", r"/health", check_health))
    cors.add(
        app.router.add_route("GET", r"/metrics", build_prometheus_metrics_handler(metric_registry))
    )
    return app


@aiotools.server_context
async def server_main(
    loop: asyncio.AbstractEventLoop,
    pidx: int,
    _args: Sequence[Any],
) -> AsyncGenerator[None, signal.Signals]:
    local_config: AgentUnifiedConfig = _args[0]

    # Start aiomonitor.
    # Port is set by config (default=50200).
    loop.set_debug(local_config.debug.asyncio)
    monitor = aiomonitor.Monitor(
        loop,
        termui_port=local_config.agent.aiomonitor_termui_port + pidx,
        webui_port=local_config.agent.aiomonitor_webui_port + pidx,
        console_enabled=False,
        hook_task_factory=local_config.debug.enhanced_aiomonitor_task_info,
    )
    Profiler(
        pyroscope_args=PyroscopeArgs(
            enabled=local_config.pyroscope.enabled,
            application_name=local_config.pyroscope.app_name,
            server_address=local_config.pyroscope.server_addr,
            sample_rate=local_config.pyroscope.sample_rate,
        )
    )

    monitor.prompt = "monitor (agent) >>> "
    monitor.console_locals["local_config"] = local_config
    aiomon_started = False
    try:
        monitor.start()
        aiomon_started = True
    except Exception as e:
        log.warning("aiomonitor could not start but skipping this error to continue", exc_info=e)

    log.info("Preparing kernel runner environments...")
    kernel_mod = importlib.import_module(
        f"ai.backend.agent.{local_config.agent.backend.value}.kernel",
    )
    krunner_volumes: Mapping[str, str] = await kernel_mod.prepare_krunner_env(
        local_config.model_dump(by_alias=True)
    )  # type: ignore
    # TODO: merge k8s branch: nfs_mount_path = local_config['baistatic']['mounted-at']
    log.info("Kernel runner environments: {}", [*krunner_volumes.keys()])
    # Update agent id and instance type if not set
    agent_updates = {}
    if not local_config.agent.id:
        agent_updates["id"] = await identity.get_instance_id()
    if not local_config.agent.instance_type:
        agent_updates["instance_type"] = await identity.get_instance_type()
    local_config.container.krunner_volumes = krunner_volumes

    if agent_updates:
        new_agent_config = local_config.agent.model_copy(update=agent_updates)
        local_config = local_config.model_copy(update={"agent": new_agent_config})

    etcd_credentials = None
    if local_config.etcd.user and local_config.etcd.password:
        etcd_credentials = {
            "user": local_config.etcd.user,
            "password": local_config.etcd.password,
        }
    scope_prefix_map = {
        ConfigScopes.GLOBAL: "",
        ConfigScopes.SGROUP: f"sgroup/{local_config.agent.scaling_group}",
        ConfigScopes.NODE: f"nodes/agents/{local_config.agent.id}",
    }
    etcd_config_data = local_config.etcd.to_dataclass()
    etcd = AsyncEtcd(
        [addr.to_legacy() for addr in etcd_config_data.addrs],
        local_config.etcd.namespace,
        scope_prefix_map,
        credentials=etcd_credentials,
    )

    rpc_addr = local_config.agent.rpc_listen_addr
    if not rpc_addr.host:
        _subnet_hint = await etcd.get("config/network/subnet/agent")
        subnet_hint = None
        if _subnet_hint is not None:
            subnet_hint = ip_network(_subnet_hint)
        log.debug("auto-detecting agent host")
        new_rpc_addr = HostPortPair(
            await identity.get_instance_ip(subnet_hint),
            rpc_addr.port,
        )
        new_agent_config = local_config.agent.model_copy(update={"rpc_listen_addr": new_rpc_addr})
        local_config = local_config.model_copy(update={"agent": new_agent_config})
    # Handle container bind-host configuration
    if not local_config.container.bind_host:
        log.debug(
            "auto-detecting `container.bind-host` from container subnet config "
            "and agent.rpc-listen-addr"
        )
        local_config.container.bind_host = await get_subnet_ip(
            etcd,
            "container",
            fallback_addr=local_config.agent.rpc_listen_addr.host,
        )
    log.info("Agent external IP: {}", local_config.agent.rpc_listen_addr.host)
    log.info("Container external IP: {}", local_config.container.bind_host)
    # Update region if not set
    if not local_config.agent.region:
        local_config.agent.region = await identity.get_instance_region()
    log.info(
        "Node ID: {0} (machine-type: {1}, host: {2})",
        local_config.agent.id,
        local_config.agent.instance_type,
        rpc_addr.host,
    )
    local_config.plugins = await etcd.get_prefix_dict("config/plugins/accelerator")

    # Start RPC server.
    global agent_instance
    agent = await AgentRPCServer.new(
        etcd,
        local_config,
        skip_detect_manager=local_config.agent.skip_manager_detection,
    )
    agent_instance = agent
    monitor.console_locals["agent"] = agent
    app = build_root_server()
    runner = web.AppRunner(app)
    await runner.setup()
    service_addr = HostPortPair(
        local_config.agent.service_addr.host, local_config.agent.service_addr.port
    )
    announce_addr = HostPortPair(
        local_config.agent.announce_addr.host, local_config.agent.announce_addr.port
    )
    ssl_ctx = None
    sd_type = ServiceDiscoveryType(local_config.service_discovery.type)

    service_discovery: ServiceDiscovery
    match sd_type:
        case ServiceDiscoveryType.ETCD:
            service_discovery = ETCDServiceDiscovery(ETCDServiceDiscoveryArgs(etcd))
        case ServiceDiscoveryType.REDIS:
            await agent.read_agent_config()
            if not local_config.redis:
                raise ConfigurationError({"server_main": "Redis runtime configuration is missing."})
            valkey_profile_target = local_config.redis.to_valkey_profile_target()
            live_valkey_target = valkey_profile_target.profile_target(RedisRole.LIVE)
            service_discovery = await RedisServiceDiscovery.create(
                args=RedisServiceDiscoveryArgs(valkey_target=live_valkey_target)
            )

    sd_loop = ServiceDiscoveryLoop(
        sd_type,
        service_discovery,
        ServiceMetadata(
            display_name=f"agent-{local_config.agent.id}",
            service_group="agent",
            version=VERSION,
            endpoint=ServiceEndpoint(
                address=str(announce_addr),
                port=announce_addr.port,
                protocol="http",
                prometheus_address=str(announce_addr),
            ),
        ),
    )
    if local_config.otel.enabled:
        meta = sd_loop.metadata
        otel_spec = OpenTelemetrySpec(
            service_id=meta.id,
            service_name=meta.service_group,
            service_version=meta.version,
            log_level=local_config.otel.log_level,
            endpoint=local_config.otel.endpoint,
        )
        BraceStyleAdapter.apply_otel(otel_spec)

    if local_config.agent.ssl_enabled:
        ssl_ctx = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
        ssl_ctx.load_cert_chain(
            str(local_config.agent.ssl_cert),
            str(local_config.agent.ssl_key),
        )
    site = web.TCPSite(
        runner,
        str(service_addr.host),
        service_addr.port,
        backlog=1024,
        reuse_port=True,
        ssl_context=ssl_ctx,
    )
    await site.start()
    log.info("started serving HTTP at {}", service_addr)

    # Run!
    try:
        async with agent:
            stop_signal = yield
            agent.mark_stop_signal(stop_signal)
    finally:
        if aiomon_started:
            monitor.close()
        sd_loop.close()


@click.group(invoke_without_command=True)
@click.option(
    "-f",
    "--config-path",
    "--config",
    type=click.Path(exists=True, dir_okay=False, path_type=Path),
    default=None,
    help="The config file path. (default: ./agent.toml and /etc/backend.ai/agent.toml)",
)
@click.option(
    "--debug",
    is_flag=True,
    help="A shortcut to set `--log-level=DEBUG`",
)
@click.option(
    "--log-level",
    type=click.Choice([*LogLevel], case_sensitive=False),
    default=LogLevel.NOTSET,
    help="Set the logging verbosity level",
)
@click.pass_context
def main(
    cli_ctx: click.Context,
    config_path: Path,
    log_level: LogLevel,
    debug: bool = False,
) -> int:
    """Start the agent service as a foreground process."""
    # Determine where to read configuration.
    try:
        raw_cfg, cfg_src_path = config.read_from_file(config_path, "agent")
    except config.ConfigurationError as e:
        print(
            "ConfigurationError: Could not read or validate the storage-proxy local config:",
            file=sys.stderr,
        )
        print(pformat(e.invalid_data), file=sys.stderr)
        raise click.Abort()

    # Override the read config with environment variables (for legacy).
    config.override_with_env(raw_cfg, ("etcd", "namespace"), "BACKEND_NAMESPACE")
    config.override_with_env(raw_cfg, ("etcd", "addr"), "BACKEND_ETCD_ADDR")
    config.override_with_env(raw_cfg, ("etcd", "user"), "BACKEND_ETCD_USER")
    config.override_with_env(raw_cfg, ("etcd", "password"), "BACKEND_ETCD_PASSWORD")
    config.override_with_env(
        raw_cfg, ("agent", "rpc-listen-addr", "host"), "BACKEND_AGENT_HOST_OVERRIDE"
    )
    config.override_with_env(raw_cfg, ("agent", "rpc-listen-addr", "port"), "BACKEND_AGENT_PORT")
    config.override_with_env(raw_cfg, ("agent", "pid-file"), "BACKEND_PID_FILE")
    config.override_with_env(raw_cfg, ("container", "port-range"), "BACKEND_CONTAINER_PORT_RANGE")
    config.override_with_env(raw_cfg, ("container", "bind-host"), "BACKEND_BIND_HOST_OVERRIDE")
    config.override_with_env(raw_cfg, ("container", "sandbox-type"), "BACKEND_SANDBOX_TYPE")
    config.override_with_env(raw_cfg, ("container", "scratch-root"), "BACKEND_SCRATCH_ROOT")

    # Validate and fill configurations
    # (allow_extra will make configs to be forward-copmatible)
    try:
        server_config = AgentUnifiedConfig.model_validate(raw_cfg)
        if debug:
            log_level = LogLevel.DEBUG
        server_config.debug.enabled = log_level == LogLevel.DEBUG
        if log_level != LogLevel.NOTSET:
            server_config.logging.level = log_level
            server_config.logging.pkg_ns["ai.backend"] = log_level
        if server_config.agent.backend == AgentBackend.KUBERNETES:
            if server_config.container.scratch_type == "k8s-nfs" and (
                server_config.container.scratch_nfs_address is None
                or server_config.container.scratch_nfs_options is None
            ):
                raise ValueError(
                    "scratch-nfs-address and scratch-nfs-options are required for k8s-nfs"
                )
        elif server_config.agent.backend == AgentBackend.DOCKER:
            DockerExtraConfig.model_validate(raw_cfg.get("container", {}))

        if server_config.debug.enabled:
            print("== Agent configuration ==")
            pprint(server_config.model_dump(by_alias=True))
    except Exception as e:
        print("ConfigurationError: Validation of agent local config has failed:", file=sys.stderr)
        print(str(e), file=sys.stderr)
        raise click.Abort()

    # FIXME: Remove this after ARM64 support lands on Jail
    current_arch = get_arch_name()
    if (
        server_config.container.sandbox_type == ContainerSandboxType.JAIL
        and current_arch != "x86_64"
    ):
        print(f"ConfigurationError: Jail sandbox is not supported on architecture {current_arch}")
        raise click.Abort()

    rpc_host = server_config.agent.rpc_listen_addr.host
    if isinstance(rpc_host, (IPv4Address, IPv6Address)) and (
        rpc_host.is_unspecified or rpc_host.is_link_local
    ):
        print(
            "ConfigurationError: "
            "Cannot use link-local or unspecified IP address as the RPC listening host.",
            file=sys.stderr,
        )
        raise click.Abort()

    if os.getuid() != 0 and server_config.container.stats_type == StatModes.CGROUP:
        print(
            "Cannot use cgroup statistics collection mode unless the agent runs as root.",
            file=sys.stderr,
        )
        raise click.Abort()

    if os.getuid() != 0 and server_config.container.scratch_type == ScratchType.HOSTFILE:
        print(
            "Cannot use hostfile scratch type unless the agent runs as root.",
            file=sys.stderr,
        )
        raise click.Abort()

    if cli_ctx.invoked_subcommand is None:
        if server_config.debug.coredump.enabled:
            if not sys.platform.startswith("linux"):
                print(
                    "ConfigurationError: Storing container coredumps is only supported in Linux.",
                    file=sys.stderr,
                )
                raise click.Abort()
            core_pattern = Path("/proc/sys/kernel/core_pattern").read_text().strip()
            if core_pattern.startswith("|") or not core_pattern.startswith("/"):
                print(
                    "ConfigurationError: "
                    "/proc/sys/kernel/core_pattern must be an absolute path "
                    "to enable container coredumps.",
                    file=sys.stderr,
                )
                raise click.Abort()
            server_config.debug.coredump.set_core_path(Path(core_pattern).parent)

        server_config.agent.pid_file.write_text(str(os.getpid()))
        image_commit_path = server_config.agent.image_commit_path
        image_commit_path.mkdir(parents=True, exist_ok=True)
        ipc_base_path = server_config.agent.ipc_base_path
        log_sockpath = ipc_base_path / f"agent-logger-{os.getpid()}.sock"
        log_sockpath.parent.mkdir(parents=True, exist_ok=True)
        log_endpoint = f"ipc://{log_sockpath}"
        try:
            logger = Logger(
                server_config.logging,
                is_master=True,
                log_endpoint=log_endpoint,
                msgpack_options={
                    "pack_opts": msgpack.DEFAULT_PACK_OPTS,
                    "unpack_opts": msgpack.DEFAULT_UNPACK_OPTS,
                },
            )
            with logger:
                ns = server_config.etcd.namespace
                setproctitle(f"backend.ai: agent {ns}")
                log.info("Backend.AI Agent {0}", VERSION)
                log.info("runtime: {0}", utils.env_info())

                log_config = logging.getLogger("ai.backend.agent.config")
                if log_level == "DEBUG":
                    log_config.debug("debug mode enabled.")
                if server_config.agent.event_loop == EventLoopType.UVLOOP:
                    import uvloop

                    uvloop.install()
                    log.info("Using uvloop as the event loop backend")
                aiotools.start_server(
                    server_main_logwrapper,
                    num_workers=1,
                    args=(server_config, log_endpoint),
                    wait_timeout=5.0,
                )
                log.info("exit.")
        finally:
            if server_config.agent.pid_file.is_file():
                # check is_file() to prevent deleting /dev/null!
                server_config.agent.pid_file.unlink()
    else:
        # Click is going to invoke a subcommand.
        pass
    return 0


if __name__ == "__main__":
    sys.exit(main())



import logging
from collections.abc import Sequence
from http import HTTPStatus
from typing import Any, List, Mapping, MutableMapping, cast
from uuid import UUID

import attr
from aiodocker.docker import Docker
from aiohttp import web
from aiohttp.typedefs import Handler, Middleware

from ai.backend.agent.config.unified import AgentUnifiedConfig
from ai.backend.agent.docker.kernel import prepare_kernel_metadata_uri_handling
from ai.backend.agent.kernel import AbstractKernel
from ai.backend.agent.utils import closing_async
from ai.backend.common.docker import LabelName
from ai.backend.common.etcd import AsyncEtcd
from ai.backend.common.json import dump_json_str
from ai.backend.common.plugin import BasePluginContext
from ai.backend.common.types import KernelId, aobject
from ai.backend.logging import BraceStyleAdapter

from .plugin import MetadataPlugin
from .root import ContainerMetadataPlugin

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


class MetadataPluginContext(BasePluginContext[MetadataPlugin]):
    plugin_group = "backendai_metadata_app_v20"


class BaseContext:
    pass


@attr.s(slots=True, auto_attribs=True, init=False)
class RootContext(BaseContext):
    local_config: Mapping[str, Any]
    etcd: AsyncEtcd
    metadata_plugin_ctx: MetadataPluginContext


async def on_prepare(request: web.Request, response: web.StreamResponse) -> None:
    response.headers["Server"] = "BackendAI"


@web.middleware
async def container_resolver_middleware(
    request: web.Request, handler: Handler
) -> web.StreamResponse:
    if (
        request.headers.get("X-Forwarded-For") is not None
        and request.app["docker-mode"] == "linuxkit"
    ):
        container_ip = request.headers["X-Forwarded-For"]
    elif remote_ip := request.remote:
        container_ip = remote_ip
    else:
        return web.Response(status=HTTPStatus.FORBIDDEN)
    async with closing_async(Docker()) as docker:
        filters = dump_json_str({
            "label": [LabelName.KERNEL_ID],
            "network": ["bridge"],
            "status": ["running"],
        })
        containers = await docker.containers.list(filters=filters)
    target_container = list(
        filter(
            lambda x: x["NetworkSettings"]["Networks"].get("bridge", {}).get("IPAddress")
            == container_ip,
            containers,
        )
    )

    if len(target_container) == 0:
        return web.Response(status=HTTPStatus.FORBIDDEN)
    request["container-ip"] = container_ip
    request["container"] = target_container[0]
    request["kernel"] = request.app["kernel-registry"].get(
        UUID(target_container[0]["Labels"][LabelName.KERNEL_ID])
    )
    return await handler(request)


async def list_versions(request: web.Request) -> web.Response:
    return web.Response(body="latest/")


class MetadataServer(aobject):
    app: web.Application
    runner: web.AppRunner
    route_structure: MutableMapping[str, Any]
    loaded_apps: List[str]

    def __init__(
        self,
        local_config: AgentUnifiedConfig,
        etcd: AsyncEtcd,
        kernel_registry: Mapping[KernelId, AbstractKernel],
    ) -> None:
        app = web.Application(
            middlewares=[
                container_resolver_middleware,
                self.route_structure_fallback_middleware,
            ],
        )
        app["_root.context"] = RootContext()
        app["_root.context"].local_config = local_config
        app["_root.context"].etcd = etcd
        app["kernel-registry"] = kernel_registry
        self.app = app
        self.loaded_apps = []
        self.route_structure = {"latest": {"extension": {}}}

    async def __ainit__(self):
        local_config = cast(AgentUnifiedConfig, self.app["_root.context"].local_config)
        await prepare_kernel_metadata_uri_handling(local_config)
        self.app["docker-mode"] = local_config.agent.docker_mode
        log.info("Loading metadata plugin: meta-data")
        metadata_plugin = ContainerMetadataPlugin({}, local_config)
        await metadata_plugin.init(None)
        metadata_app, global_middlewares, route_structures = await metadata_plugin.create_app()
        self._init_subapp(
            "meta-data",
            self.app,
            metadata_app,
            global_middlewares,
            route_structures,
            is_extension=False,
        )
        self.app.router.add_route("GET", "/", list_versions)
        self.app.router.add_route("GET", "/{version}", self.list_available_apps)

    async def list_available_apps(self, request: web.Request) -> web.Response:
        return web.Response(body="\n".join([x + "/" for x in self.loaded_apps]))

    @web.middleware
    async def route_structure_fallback_middleware(
        self, request: web.Request, handler: Handler
    ) -> web.StreamResponse:
        try:
            return await handler(request)
        except web.HTTPNotFound:
            components = request.path.split("/")
            if len(components) > 0 and components[0] == "":
                components = components[1:]
            if len(components) > 0 and components[-1] == "":
                components = components[:-1]
            structure_pointer = self.route_structure
            for component in components:
                if structure_pointer.get(component) is None:
                    raise web.HTTPNotFound
                structure_pointer = structure_pointer[component]
            resources = []
            for k, v in structure_pointer.items():
                if isinstance(v, dict):
                    resources.append(k + "/")
                else:
                    resources.append(k)
            resources.sort()
            return web.Response(body="\n".join(resources))

    def _init_subapp(
        self,
        pkg_name: str,
        root_app: web.Application,
        subapp: web.Application,
        global_middlewares: Sequence[Middleware],
        route_structure: Mapping[str, Any],
        is_extension: bool = True,
    ) -> None:
        subapp.on_response_prepare.append(on_prepare)

        async def _set_root_ctx(subapp: web.Application):
            # Allow subapp's access to the root app properties.
            # These are the public APIs exposed to plugins as well.
            subapp["_root.context"] = root_app["_root.context"]

        # We must copy the public interface prior to all user-defined startup signal handlers.
        subapp.on_startup.insert(0, _set_root_ctx)
        prefix = subapp.get("prefix", pkg_name.split(".")[-1].replace("_", "-"))
        if is_extension:
            self.route_structure["latest"]["extension"][prefix] = route_structure
            prefix = "extension/" + prefix
        else:
            self.route_structure["latest"][prefix] = route_structure
        root_app.add_subapp("/latest/" + prefix, subapp)
        root_app.middlewares.extend(global_middlewares)
        self.loaded_apps.append(prefix)

    async def load_metadata_plugins(self) -> None:
        root_ctx = self.app["_root.context"]
        plugin_ctx = MetadataPluginContext(root_ctx.etcd, root_ctx.local_config)
        await plugin_ctx.init()
        root_ctx.metadata_plugin_ctx = plugin_ctx
        log.debug("Available metadata plugins: {}", plugin_ctx.plugins)
        for plugin_name, plugin_instance in plugin_ctx.plugins.items():
            log.info("Loading metadata plugin: {0}", plugin_name)
            subapp, global_middlewares, route_structure = await plugin_instance.create_app()
            self._init_subapp(plugin_name, self.app, subapp, global_middlewares, route_structure)

    async def start_server(self) -> None:
        await self.load_metadata_plugins()
        metadata_server_runner = web.AppRunner(self.app)
        await metadata_server_runner.setup()
        local_config: AgentUnifiedConfig = self.app["_root.context"].local_config
        site = web.TCPSite(
            metadata_server_runner,
            local_config.agent.metadata_server_bind_host,
            local_config.agent.metadata_server_port,
        )
        self.runner = metadata_server_runner
        await site.start()

    async def cleanup(self) -> None:
        plugin_context = self.app["_root.context"].metadata_plugin_ctx
        await self.runner.cleanup()
        await self.app.shutdown()
        await plugin_context.cleanup()



"""
Kernel object creation stage for kernel lifecycle.

This stage handles creation of the final DockerKernel object.
"""

from dataclasses import dataclass
from typing import Any, Mapping, Sequence, override

from ai.backend.agent.docker.kernel import DockerKernel
from ai.backend.agent.resources import KernelResourceSpec
from ai.backend.agent.types import KernelOwnershipData
from ai.backend.common.docker import ImageRef
from ai.backend.common.stage.types import ArgsSpecGenerator, Provisioner, ProvisionStage
from ai.backend.common.types import ClusterInfo, ServicePort


@dataclass
class KernelObjectSpec:
    ownership_data: KernelOwnershipData
    kernel_config: Mapping[str, Any]
    image_ref: ImageRef
    kspec_version: int
    cluster_info: ClusterInfo
    agent_config: Mapping[str, Any]
    service_ports: Sequence[ServicePort]
    resource_spec: KernelResourceSpec
    environ: Mapping[str, str]


class KernelObjectSpecGenerator(ArgsSpecGenerator[KernelObjectSpec]):
    pass


@dataclass
class KernelObjectResult:
    kernel: DockerKernel


class KernelObjectProvisioner(Provisioner[KernelObjectSpec, KernelObjectResult]):
    """
    Provisioner for kernel object creation.

    Creates the final DockerKernel instance with all prepared configurations.
    """

    @property
    @override
    def name(self) -> str:
        return "docker-kernel-object"

    @override
    async def setup(self, spec: KernelObjectSpec) -> KernelObjectResult:
        # Extract network configuration
        network_mode = spec.cluster_info["network_config"].get("mode", "bridge")

        # Create DockerKernel instance
        kernel = DockerKernel(
            ownership_data=spec.ownership_data,
            network_id=spec.kernel_config["network_id"],
            image=spec.image_ref,
            version=spec.kspec_version,
            network_driver=network_mode,
            agent_config=spec.agent_config,
            service_ports=spec.service_ports,
            resource_spec=spec.resource_spec,
            environ=spec.environ,
            data={},  # Additional data can be added here if needed
        )

        return KernelObjectResult(kernel=kernel)

    @override
    async def teardown(self, resource: KernelObjectResult) -> None:
        # No cleanup needed for kernel object
        pass


class KernelObjectStage(ProvisionStage[KernelObjectSpec, KernelObjectResult]):
    """
    Stage for creating the final DockerKernel object.
    """

    pass
